from numpy import array
from numpy import linspace
from numpy import ones
from numpy import zeros

from compas_tno.shapes import MeshDos
from compas_tno.shapes import rectangular_topology

import math


def pointed_vault_heightfields_proxy(xy_span=[[0.0, 10.0], [0.0, 10.0]], thk=0.5, discretisation=[20, 20], hc=8.0, he=None, hm=None, tol=10e-6, t=0.0, *args, **kwargs):
    """Set pointed cross vault heights through a proxy

    Parameters
    ----------
    xy_span : list, optional
        [description], by default [[0.0, 10.0], [0.0, 10.0]]
    thk : float, optional
        [description], by default 0.5
    discretisation : list, optional
        [description], by default [10, 10]
    hc : float, optional
        Height in the middle point of the vault, by default 8.0
    he : [float, float, float, float], optional
        Height of the opening mid-span for each of the quadrants, by default None
    hm : [float, float, float, float], optional
        Height of each quadrant center (spadrel), by default None
    tol : float, optional
        Tolerance, by default 10e-6
    t : float, optional
        Parameter for lower bound in nodes in the boundary, by default 0.0

    Returns
    -------
    intradosdata
        The data of MeshDos for the intrados of the shape
    extradosdata
        The data of MeshDos for the extrados of the shape
    middledata
        The data of MeshDos for the middle of the shape

    Notes
    ----------------------
    Position of the quadrants is as in the schema below:

        Q3
    Q2      Q1
        Q4
    """
    intrados, extrados, middle = pointed_vault_heightfields(xy_span=xy_span, thk=thk, discretisation=discretisation, hc=hc, he=he, hm=hm, tol=tol, t=t)
    return intrados.to_data(), extrados.to_data(), middle.to_data()


def pointed_vault_heightfields(xy_span=[[0.0, 10.0], [0.0, 10.0]], thk=0.5, discretisation=[10, 10], hc=8.0, he=None, hm=None,  t=0.0, tol=0.00):
    """Set pointed cross vault heights

    Parameters
    ----------
    xy_span : list, optional
        [description], by default [[0.0, 10.0], [0.0, 10.0]]
    thk : float, optional
        [description], by default 0.5
    discretisation : list, optional
        [description], by default [10, 10]
    hc : float, optional
        Height in the middle point of the vault, by default 8.0
    he : [float, float, float, float], optional
        Height of the opening mid-span for each of the quadrants, by default None
    hm : [float, float, float, float], optional
        Height of each quadrant center (spadrel), by default None
    t : float, optional
        Parameter for lower bound in nodes in the boundary, by default 0.0
    tol : float, optional
        Tolerance, by default 10e-6

    Returns
    -------
    intrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the intrados of the shape
    extrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the extrados of the shape
    middle : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the middle of the shape

    Notes
    ----------------------
    Position of the quadrants is as in the schema below:

        Q3
    Q2      Q1
        Q4
    """

    if isinstance(discretisation, int):
        discretisation = [discretisation, discretisation]

    y1 = xy_span[1][1]
    y0 = xy_span[1][0]
    x1 = xy_span[0][1]
    x0 = xy_span[0][0]

    density_x = discretisation[0]
    density_y = discretisation[1]
    x = linspace(x0, x1, num=density_x+1, endpoint=True)  # arange(x0, x1 + dx/density_x, dx/density_x)
    y = linspace(y0, y1, num=density_y+1, endpoint=True)  # arange(y0, y1 + dy/density_y, dy/density_y)

    xi, yi, faces_i = rectangular_topology(x, y)

    zt = pointed_vault_middle_update(xi, yi, t, xy_span=xy_span, hc=hc, he=he, hm=hm, tol=tol)
    xyzt = array([xi, yi, zt.flatten()]).transpose()
    middle = MeshDos.from_vertices_and_faces(xyzt, faces_i)

    zub, zlb = pointed_vault_ub_lb_update(xi, yi, thk, t, xy_span=xy_span, hc=hc, he=he, hm=hm, tol=tol)
    xyzub = array([xi, yi, zub.flatten()]).transpose()
    xyzlb = array([xi, yi, zlb.flatten()]).transpose()
    extrados = MeshDos.from_vertices_and_faces(xyzub, faces_i)
    intrados = MeshDos.from_vertices_and_faces(xyzlb, faces_i)

    return intrados, extrados, middle


def pointed_vault_middle_update(x, y, thk, xy_span=[[0.0, 10.0], [0.0, 10.0]], hc=8.0, he=None, hm=None, tol=1e-6):
    """Update middle of a pointed vault based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    hc : float, optional
        Height in the middle point of the vault, by default 8.0
    he : [float, float, float, float], optional
        Height of the opening mid-span for each of the quadrants, by default None
    hm : [float, float, float, float], optional
        Height of each quadrant center (spadrel), by default None
    tol : float, optional
        Tolerance, by default 10e-6

    Returns
    -------
    middle : array
        Values of the middle surface in the points
    """

    y1 = xy_span[1][1]
    y0 = xy_span[1][0]
    x1 = xy_span[0][1]
    x0 = xy_span[0][0]
    lx = x1 - x0
    ly = y1 - y0

    if he and hm is None:
        h1, k1, r1 = _circle_3points_xy([x0, he[1]], [(x1+x0)/2, hc], [x1, he[0]])
        h2, k2, r2 = h1, k1, r1
        h3, k3, r3 = _circle_3points_xy([y0, he[3]], [(y1+y0)/2, hc], [y1, he[2]])
        h4, k4, r4 = h3, k3, r3
    elif hm and he:
        h1, k1, r1 = _circle_3points_xy([(x1+x0)/2, hc], [3*(x1+x0)/4, hm[0]], [x1, he[0]])
        h2, k2, r2 = _circle_3points_xy([(x1+x0)/2, hc], [1*(x1+x0)/4, hm[1]], [x0, he[1]])
        h3, k3, r3 = _circle_3points_xy([(y1+y0)/2, hc], [3*(y1+y0)/4, hm[2]], [y1, he[2]])
        h4, k4, r4 = _circle_3points_xy([(y1+y0)/2, hc], [1*(y1+y0)/4, hm[3]], [y0, he[3]])

    middle = zeros((len(x), 1))

    for i in range(len(x)):
        xi, yi = x[i], y[i]

        if yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q1
            # Equation (xi - hx) ** 2 + (hi - kx) ** 2 = rx **2 to find the height of the pointed part (middle of quadrant) with that height one find the equivalent radius
            if he:
                hi = k1 + math.sqrt(r1 ** 2 - (xi - h1) ** 2)
            else:
                hi = hc
            ri = _find_r_given_h_l(hi, ly)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            if yi <= (y1 + y0)/2:
                zi = _sqrt((ri)**2 - (yi-(y0+ri))**2)
            else:
                zi = _sqrt((ri)**2 - (yi-(y1-ri))**2)

        elif yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q3
            # Equation (xi - hy) ** 2 + (hi - ky) ** 2 = ry **2 to find the height of the pointed part (middle of quadrant) with that height one find the equivalent radius
            if he:
                hi = k3 + math.sqrt(r3 ** 2 - (yi - h3) ** 2)
            else:
                hi = hc
            ri = _find_r_given_h_l(hi, lx)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            if xi <= (x0 + x1)/2:
                zi = _sqrt((ri)**2 - (xi-(x0+ri))**2)
            else:
                zi = _sqrt((ri)**2 - (xi-(x1-ri))**2)

        elif yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q2
            if he:
                hi = k2 + math.sqrt(r2 ** 2 - (xi - h2) ** 2)
            else:
                hi = hc
            ri = _find_r_given_h_l(hi, ly)
            if yi <= (y1 + y0)/2:
                zi = _sqrt((ri)**2 - (yi-(y0+ri))**2)
            else:
                zi = _sqrt((ri)**2 - (yi-(y1-ri))**2)

        elif yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q4
            if he:
                hi = k4 + math.sqrt(r4 ** 2 - (yi - h4) ** 2)
            else:
                hi = hc
            ri = _find_r_given_h_l(hi, lx)
            if xi <= (x0 + x1)/2:
                zi = _sqrt((ri)**2 - (xi-(x0+ri))**2)
            else:
                zi = _sqrt((ri)**2 - (xi-(x1-ri))**2)

        else:
            print('Vertex did not belong to any Q. (x,y) = ({0},{1})'.format(xi, yi))

        middle[i] = zi

    return middle


def pointed_vault_ub_lb_update(x, y, thk, t, xy_span=[[0.0, 10.0], [0.0, 10.0]], hc=8.0, he=None, hm=None, tol=1e-6):
    """Update upper and lower bounds of a pointed vault based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    t : float
        Parameter for lower bound in nodes in the boundary
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    hc : float, optional
        Height in the middle point of the vault, by default 8.0
    he : [float, float, float, float], optional
        Height of the opening mid-span for each of the quadrants, by default None
    hm : [float, float, float, float], optional
        Height of each quadrant center (spadrel), by default None
    tol : float, optional
        Tolerance, by default 10e-6

    Returns
    -------
    ub : array
        Values of the upper bound in the points
    lb : array
        Values of the lower bound in the points
    """

    y1 = xy_span[1][1]
    y0 = xy_span[1][0]
    x1 = xy_span[0][1]
    x0 = xy_span[0][0]

    # y1_ub = y1 + thk/2
    # y0_ub = y0 - thk/2
    # x1_ub = x1 + thk/2
    # x0_ub = x0 - thk/2

    y1_lb = y1 - thk/2
    y0_lb = y0 + thk/2
    x1_lb = x1 - thk/2
    x0_lb = x0 + thk/2

    # lx_ub = x1_ub - x0_ub
    # ly_ub = y1_ub - y0_ub

    # lx_lb = x1_lb - x0_lb
    # ly_lb = y1_lb - y0_lb

    lx = x1 - x0
    ly = y1 - y0

    # hc_ub = hc + thk/2
    # hc_lb = hc - thk/2

    # r_x = _find_r_given_h_l(hc, x1 - x0)
    # r_y = _find_r_given_h_l(hc, y1 - y0)

    if he:
        he_ub = he.copy()
        he_lb = he.copy()
        for i in range(len(he)):
            he_ub[i] += thk/2
            he_lb[i] -= thk/2
    if hm:
        raise NotImplementedError()
        hm_ub = hm.copy()
        hm_lb = hm.copy()
        for i in range(len(hm)):
            hm_ub[i] += thk/2
            hm_lb[i] -= thk/2

    if he and hm is None:
        # h1_ub, k1_ub, r1_ub = _circle_3points_xy([x0, he_ub[1]], [(x1+x0)/2, hc_ub], [x1, he_ub[0]])
        # h2_ub, k2_ub, r2_ub = h1_ub, k1_ub, r1_ub
        # h3_ub, k3_ub, r3_ub = _circle_3points_xy([y0, he_ub[3]], [(y1+y0)/2, hc_ub], [y1, he_ub[2]])
        # h4_ub, k4_ub, r4_ub = h3_ub, k3_ub, r3_ub

        # h1_lb, k1_lb, r1_lb = _circle_3points_xy([x0, he_lb[1]], [(x1+x0)/2, hc_lb], [x1, he_lb[0]])
        # h2_lb, k2_lb, r2_lb = h1_lb, k1_lb, r1_lb
        # h3_lb, k3_lb, r3_lb = _circle_3points_xy([y0, he_lb[3]], [(y1+y0)/2, hc_lb], [y1, he_lb[2]])
        # h4_lb, k4_lb, r4_lb = h3_lb, k3_lb, r3_lb

        h1, k1, r1 = _circle_3points_xy([x0, he[1]], [(x1+x0)/2, hc], [x1, he[0]])
        h2, k2, r2 = h1, k1, r1
        h3, k3, r3 = _circle_3points_xy([y0, he[3]], [(y1+y0)/2, hc], [y1, he[2]])
        h4, k4, r4 = h3, k3, r3

    # elif hm and he:
    #     h1_ub, k1_ub, r1_ub = _circle_3points_xy([(x1+x0)/2, hc_ub], [3*(x1+x0)/4, hm_ub[0]], [x1, he_ub[0]])
    #     h2_ub, k2_ub, r2_ub = _circle_3points_xy([(x1+x0)/2, hc_ub], [1*(x1+x0)/4, hm_ub[1]], [x0, he_ub[1]])
    #     h3_ub, k3_ub, r3_ub = _circle_3points_xy([(y1+y0)/2, hc_ub], [3*(y1+y0)/4, hm_ub[2]], [y1, he_ub[2]])
    #     h4_ub, k4_ub, r4_ub = _circle_3points_xy([(y1+y0)/2, hc_ub], [1*(y1+y0)/4, hm_ub[3]], [y0, he_ub[3]])

    #     h1_lb, k1_lb, r1_lb = _circle_3points_xy([(x1+x0)/2, hc_lb], [3*(x1+x0)/4, hm_lb[0]], [x1, he_lb[0]])
    #     h2_lb, k2_lb, r2_lb = _circle_3points_xy([(x1+x0)/2, hc_lb], [1*(x1+x0)/4, hm_lb[1]], [x0, he_lb[1]])
    #     h3_lb, k3_lb, r3_lb = _circle_3points_xy([(y1+y0)/2, hc_lb], [3*(y1+y0)/4, hm_lb[2]], [y1, he_lb[2]])
    #     h4_lb, k4_lb, r4_lb = _circle_3points_xy([(y1+y0)/2, hc_lb], [1*(y1+y0)/4, hm_lb[3]], [y0, he_lb[3]])

    ub = ones((len(x), 1))
    lb = ones((len(x), 1)) * - t

    for i in range(len(x)):
        xi, yi = x[i], y[i]

        if yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q1
            if he:
                # hi_ub = k1_ub + math.sqrt(r1_ub ** 2 - (xi - h1_ub) ** 2)  # is in fact hi
                # hi_lb = k1_lb + math.sqrt(r1_lb ** 2 - (xi - h1_lb) ** 2)
                hi = k1 + math.sqrt(r1 ** 2 - (xi - h1) ** 2)
            else:
                hi = hc
                # hi_ub = hc_ub
                # hi_lb = hc_lb
            # ri_ub = _find_r_given_h_l(hi_ub, ly_ub)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_lb = _find_r_given_h_l(hi_lb, ly_lb)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri = _find_r_given_h_l(hi, ly)
            ri_ub = ri + thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri_lb = ri - thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_ub = r_y + thk/2  # This only works for pointed (he only)
            # ri_lb = r_y - thk/2  # This only works for pointed (he only)
            if yi <= (y1 + y0)/2:
                ub[i] = _sqrt((ri_ub)**2 - (yi-(y0+ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (yi-(y0+ri))**2)
            else:
                ub[i] = _sqrt((ri_ub)**2 - (yi-(y1-ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (yi-(y1-ri))**2)

        elif yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q3
            if he:
                # hi_ub = k3_ub + math.sqrt(r3_ub ** 2 - (yi - h3_ub) ** 2)
                # hi_lb = k3_lb + math.sqrt(r3_lb ** 2 - (yi - h3_lb) ** 2)
                hi = k3 + math.sqrt(r3 ** 2 - (yi - h3) ** 2)
            else:
                hi = hc
                # hi_ub = hc_ub
                # hi_lb = hc_lb
            # ri_ub = _find_r_given_h_l(hi_ub, lx_ub)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_lb = _find_r_given_h_l(hi_lb, lx_lb)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri = _find_r_given_h_l(hi, lx)
            ri_ub = ri + thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri_lb = ri - thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_ub = r_x + thk/2  # This only works for pointed (he only)
            # ri_lb = r_x - thk/2  # This only works for pointed (he only)
            if xi <= (x0 + x1)/2:
                ub[i] = _sqrt((ri_ub)**2 - (xi-(x0+ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (xi-(x0+ri))**2)
            else:
                ub[i] = _sqrt((ri_ub)**2 - (xi-(x1-ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (xi-(x1-ri))**2)

        elif yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q2
            if he:
                # hi_ub = k2_ub + math.sqrt(r2_ub ** 2 - (xi - h2_ub) ** 2)
                # hi_lb = k2_lb + math.sqrt(r2_lb ** 2 - (xi - h2_lb) ** 2)
                hi = k2 + math.sqrt(r2 ** 2 - (xi - h2) ** 2)
            else:
                hi = hc
                # hi_ub = hc_ub
                # hi_lb = hc_lb
            # ri_lb = _find_r_given_h_l(hi_lb, ly_lb)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_ub = _find_r_given_h_l(hi_ub, ly_ub)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri = _find_r_given_h_l(hi, ly)
            ri_lb = ri - thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri_ub = ri + thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_ub = r_y + thk/2  # This only works for pointed (he only)
            # ri_lb = r_y - thk/2  # This only works for pointed (he only)
            if yi <= (y1 + y0)/2:
                ub[i] = _sqrt((ri_ub)**2 - (yi-(y0+ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (yi-(y0+ri))**2)
            else:
                ub[i] = _sqrt((ri_ub)**2 - (yi-(y1-ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (yi-(y1-ri))**2)

        elif yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q4
            if he:
                # hi_ub = k4_ub + math.sqrt(r4_ub ** 2 - (yi - h4_ub) ** 2)
                # hi_lb = k4_lb + math.sqrt(r4_lb ** 2 - (yi - h4_lb) ** 2)
                hi = k4 + math.sqrt(r4 ** 2 - (yi - h4) ** 2)
            else:
                hi = hc
                # hi_ub = hc_ub
                # hi_lb = hc_lb
            # ri_ub = _find_r_given_h_l(hi_ub, lx_ub)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_lb = _find_r_given_h_l(hi_lb, lx_lb)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri = _find_r_given_h_l(hi, lx)
            ri_ub = ri + thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri_lb = ri - thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_ub = r_x + thk/2  # This only works for pointed (he only)
            # ri_lb = r_x - thk/2  # This only works for pointed (he only)
            if xi <= (x0 + x1)/2:
                ub[i] = _sqrt((ri_ub)**2 - (xi-(x0+ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (xi-(x0+ri))**2)
            else:
                ub[i] = _sqrt((ri_ub)**2 - (xi-(x1-ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (xi-(x1-ri))**2)
        else:
            print('Vertex did not belong to any Q. (x,y) = ({0},{1})'.format(xi, yi))

        if ((yi) > y1_lb and ((xi) > x1_lb or (xi) < x0_lb)) or ((yi) < y0_lb and ((xi) > x1_lb or (xi) < x0_lb)):
            lb[i] = - 1*t

    return ub, lb


def pointed_vault_dub_dlb(x, y, thk, t, xy_span=[[0.0, 10.0], [0.0, 10.0]], hc=8.0, he=None, hm=None, tol=1e-6):
    """Computes the sensitivities of upper and lower bounds in the x, y coordinates and thickness specified.

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    t : float
        Parameter for lower bound in nodes in the boundary
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    hc : float, optional
        Height in the middle point of the vault, by default 8.0
    he : [float, float, float, float], optional
        Height of the opening mid-span for each of the quadrants, by default None
    hm : [float, float, float, float], optional
        Height of each quadrant center (spadrel), by default None
    tol : float, optional
        Tolerance, by default 10e-6

    Returns
    -------
    dub : array
        Values of the sensitivities for the upper bound in the points
    dlb : array
        Values of the sensitivities for the lower bound in the points
    """

    y1 = xy_span[1][1]
    y0 = xy_span[1][0]
    x1 = xy_span[0][1]
    x0 = xy_span[0][0]

    # y1_ub = y1 + thk/2
    # y0_ub = y0 - thk/2
    # x1_ub = x1 + thk/2
    # x0_ub = x0 - thk/2

    y1_lb = y1 - thk/2
    y0_lb = y0 + thk/2
    x1_lb = x1 - thk/2
    x0_lb = x0 + thk/2

    # lx_ub = x1_ub - x0_ub
    # ly_ub = y1_ub - y0_ub

    # lx_lb = x1_lb - x0_lb
    # ly_lb = y1_lb - y0_lb

    lx = x1 - x0
    ly = y1 - y0

    # hc_ub = hc + thk/2
    # hc_lb = hc - thk/2

    # r_x = _find_r_given_h_l(hc, x1 - x0)
    # r_y = _find_r_given_h_l(hc, y1 - y0)

    if he:
        he_ub = he.copy()
        he_lb = he.copy()
        for i in range(len(he)):
            he_ub[i] += thk/2
            he_lb[i] -= thk/2
    if hm:
        raise NotImplementedError()
        hm_ub = hm.copy()
        hm_lb = hm.copy()
        for i in range(len(hm)):
            hm_ub[i] += thk/2
            hm_lb[i] -= thk/2

    if he and hm is None:
        # h1_ub, k1_ub, r1_ub = _circle_3points_xy([x0, he_ub[1]], [(x1+x0)/2, hc_ub], [x1, he_ub[0]])
        # h2_ub, k2_ub, r2_ub = h1_ub, k1_ub, r1_ub
        # h3_ub, k3_ub, r3_ub = _circle_3points_xy([y0, he_ub[3]], [(y1+y0)/2, hc_ub], [y1, he_ub[2]])
        # h4_ub, k4_ub, r4_ub = h3_ub, k3_ub, r3_ub

        # h1_lb, k1_lb, r1_lb = _circle_3points_xy([x0, he_lb[1]], [(x1+x0)/2, hc_lb], [x1, he_lb[0]])
        # h2_lb, k2_lb, r2_lb = h1_lb, k1_lb, r1_lb
        # h3_lb, k3_lb, r3_lb = _circle_3points_xy([y0, he_lb[3]], [(y1+y0)/2, hc_lb], [y1, he_lb[2]])
        # h4_lb, k4_lb, r4_lb = h3_lb, k3_lb, r3_lb

        h1, k1, r1 = _circle_3points_xy([x0, he[1]], [(x1+x0)/2, hc], [x1, he[0]])
        h2, k2, r2 = h1, k1, r1
        h3, k3, r3 = _circle_3points_xy([y0, he[3]], [(y1+y0)/2, hc], [y1, he[2]])
        h4, k4, r4 = h3, k3, r3

    # elif hm and he:
    #     h1_ub, k1_ub, r1_ub = _circle_3points_xy([(x1+x0)/2, hc_ub], [3*(x1+x0)/4, hm_ub[0]], [x1, he_ub[0]])
    #     h2_ub, k2_ub, r2_ub = _circle_3points_xy([(x1+x0)/2, hc_ub], [1*(x1+x0)/4, hm_ub[1]], [x0, he_ub[1]])
    #     h3_ub, k3_ub, r3_ub = _circle_3points_xy([(y1+y0)/2, hc_ub], [3*(y1+y0)/4, hm_ub[2]], [y1, he_ub[2]])
    #     h4_ub, k4_ub, r4_ub = _circle_3points_xy([(y1+y0)/2, hc_ub], [1*(y1+y0)/4, hm_ub[3]], [y0, he_ub[3]])

    #     h1_lb, k1_lb, r1_lb = _circle_3points_xy([(x1+x0)/2, hc_lb], [3*(x1+x0)/4, hm_lb[0]], [x1, he_lb[0]])
    #     h2_lb, k2_lb, r2_lb = _circle_3points_xy([(x1+x0)/2, hc_lb], [1*(x1+x0)/4, hm_lb[1]], [x0, he_lb[1]])
    #     h3_lb, k3_lb, r3_lb = _circle_3points_xy([(y1+y0)/2, hc_lb], [3*(y1+y0)/4, hm_lb[2]], [y1, he_lb[2]])
    #     h4_lb, k4_lb, r4_lb = _circle_3points_xy([(y1+y0)/2, hc_lb], [1*(y1+y0)/4, hm_lb[3]], [y0, he_lb[3]])

    ub = ones((len(x), 1))
    lb = ones((len(x), 1)) * - t
    dub = zeros((len(x), 1))
    dlb = zeros((len(x), 1))

    for i in range(len(x)):
        xi, yi = x[i], y[i]

        if yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q1
            if he:
                # hi_ub = k1_ub + math.sqrt(r1_ub ** 2 - (xi - h1_ub) ** 2)  # is in fact hi
                # hi_lb = k1_lb + math.sqrt(r1_lb ** 2 - (xi - h1_lb) ** 2)
                hi = k1 + math.sqrt(r1 ** 2 - (xi - h1) ** 2)
            else:
                hi = hc
                # hi_ub = hc_ub
                # hi_lb = hc_lb
            # ri_ub = _find_r_given_h_l(hi_ub, ly_ub)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_lb = _find_r_given_h_l(hi_lb, ly_lb)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri = _find_r_given_h_l(hi, ly)
            ri_ub = ri + thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri_lb = ri - thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_ub = r_y + thk/2  # This only works for pointed (he only)
            # ri_lb = r_y - thk/2  # This only works for pointed (he only)
            if yi <= (y1 + y0)/2:
                ub[i] = _sqrt((ri_ub)**2 - (yi-(y0+ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (yi-(y0+ri))**2)
            else:
                ub[i] = _sqrt((ri_ub)**2 - (yi-(y1-ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (yi-(y1-ri))**2)
            dub[i] = 1/2 * ri_ub/ub[i]
            dlb[i] = - 1/2 * ri_lb/lb[i]

        elif yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q3
            if he:
                # hi_ub = k3_ub + math.sqrt(r3_ub ** 2 - (yi - h3_ub) ** 2)
                # hi_lb = k3_lb + math.sqrt(r3_lb ** 2 - (yi - h3_lb) ** 2)
                hi = k3 + math.sqrt(r3 ** 2 - (yi - h3) ** 2)
            else:
                hi = hc
                # hi_ub = hc_ub
                # hi_lb = hc_lb
            # ri_ub = _find_r_given_h_l(hi_ub, lx_ub)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_lb = _find_r_given_h_l(hi_lb, lx_lb)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri = _find_r_given_h_l(hi, lx)
            ri_ub = ri + thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri_lb = ri - thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_ub = r_x + thk/2  # This only works for pointed (he only)
            # ri_lb = r_x - thk/2  # This only works for pointed (he only)
            if xi <= (x0 + x1)/2:
                ub[i] = _sqrt((ri_ub)**2 - (xi-(x0+ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (xi-(x0+ri))**2)
            else:
                ub[i] = _sqrt((ri_ub)**2 - (xi-(x1-ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (xi-(x1-ri))**2)
            dub[i] = 1/2 * ri_ub/ub[i]
            dlb[i] = - 1/2 * ri_lb/lb[i]

        elif yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q2
            if he:
                # hi_ub = k2_ub + math.sqrt(r2_ub ** 2 - (xi - h2_ub) ** 2)
                # hi_lb = k2_lb + math.sqrt(r2_lb ** 2 - (xi - h2_lb) ** 2)
                hi = k2 + math.sqrt(r2 ** 2 - (xi - h2) ** 2)
            else:
                hi = hc
                # hi_ub = hc_ub
                # hi_lb = hc_lb
            # ri_lb = _find_r_given_h_l(hi_lb, ly_lb)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_ub = _find_r_given_h_l(hi_ub, ly_ub)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri = _find_r_given_h_l(hi, ly)
            ri_lb = ri - thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri_ub = ri + thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_ub = r_y + thk/2  # This only works for pointed (he only)
            # ri_lb = r_y - thk/2  # This only works for pointed (he only)
            if yi <= (y1 + y0)/2:
                ub[i] = _sqrt((ri_ub)**2 - (yi-(y0+ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (yi-(y0+ri))**2)
            else:
                ub[i] = _sqrt((ri_ub)**2 - (yi-(y1-ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (yi-(y1-ri))**2)
            dub[i] = 1/2 * ri_ub/ub[i]
            dlb[i] = - 1/2 * ri_lb/lb[i]

        elif yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q4
            if he:
                # hi_ub = k4_ub + math.sqrt(r4_ub ** 2 - (yi - h4_ub) ** 2)
                # hi_lb = k4_lb + math.sqrt(r4_lb ** 2 - (yi - h4_lb) ** 2)
                hi = k4 + math.sqrt(r4 ** 2 - (yi - h4) ** 2)
            else:
                hi = hc
                # hi_ub = hc_ub
                # hi_lb = hc_lb
            # ri_ub = _find_r_given_h_l(hi_ub, lx_ub)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_lb = _find_r_given_h_l(hi_lb, lx_lb)  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri = _find_r_given_h_l(hi, lx)
            ri_ub = ri + thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            ri_lb = ri - thk/2  # This in the equation ri ** 2 =  (xi - xc_) ** 2 + (zi - zc_) ** 2  -> zc = 0.0 and xc_ = (x0 + x1)/2
            # ri_ub = r_x + thk/2  # This only works for pointed (he only)
            # ri_lb = r_x - thk/2  # This only works for pointed (he only)
            if xi <= (x0 + x1)/2:
                ub[i] = _sqrt((ri_ub)**2 - (xi-(x0+ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (xi-(x0+ri))**2)
            else:
                ub[i] = _sqrt((ri_ub)**2 - (xi-(x1-ri))**2)
                lb[i] = _sqrt((ri_lb)**2 - (xi-(x1-ri))**2)
            dub[i] = 1/2 * ri_ub/ub[i]
            dlb[i] = - 1/2 * ri_lb/lb[i]
        else:
            print('Vertex did not belong to any Q. (x,y) = ({0},{1})'.format(xi, yi))

        if ((yi) > y1_lb and ((xi) > x1_lb or (xi) < x0_lb)) or ((yi) < y0_lb and ((xi) > x1_lb or (xi) < x0_lb)):
            lb[i] = - 1*t
            dlb[i] = 0.0

    return dub, dlb  # ub, lb


def _find_r_given_h_l(h, length):

    r = h**2/length + length/4

    return r


def _circle_3points_xy(p1, p2, p3):

    x1 = p1[0]
    z1 = p1[1]
    x2 = p2[0]
    z2 = p2[1]
    x3 = p3[0]
    z3 = p3[1]

    x12 = x1 - x2
    x13 = x1 - x3
    z12 = z1 - z2
    z13 = z1 - z3
    z31 = z3 - z1
    z21 = z2 - z1
    x31 = x3 - x1
    x21 = x2 - x1

    sx13 = x1**2 - x3**2
    sz13 = z1**2 - z3**2
    sx21 = x2**2 - x1**2
    sz21 = z2**2 - z1**2

    f = ((sx13) * (x12) + (sz13) * (x12) + (sx21) * (x13) + (sz21) * (x13)) / (2 * ((z31) * (x12) - (z21) * (x13)))
    g = ((sx13) * (z12) + (sz13) * (z12) + (sx21) * (z13) + (sz21) * (z13)) / (2 * ((x31) * (z12) - (x21) * (z13)))
    c = - x1 ** 2 - z1 ** 2 - 2 * g * x1 - 2 * f * z1
    h = - g
    k = - f
    r2 = h * h + k * k - c
    r = math.sqrt(r2)

    # print('h: ', h, 'k: ', k, 'r: ', r)

    return h, k, r


def _sqrt(x):
    try:
        sqrt_x = math.sqrt(x)
    except BaseException:
        if x > -10e4:
            sqrt_x = math.sqrt(abs(x))
        else:
            sqrt_x = 0.0
            print('Problems to sqrt: ', x)
    return sqrt_x



# from compas_tno.diagrams.form import energy
# from compas_tno.diagrams.form import loadpath
# from compas_tno.diagrams.form import adapt_tna
# from compas_tno.diagrams.form import evaluate_a

# from scipy import tensordot
# from scipy.optimize import nnls

from sympy import tensorproduct

from compas.numerical import normrow
from compas.geometry import normalize_vector
from compas.geometry import is_ccw_xy

from numpy import array
from numpy import cross
from numpy import dot
from numpy import zeros
from numpy import int8


__all__ = [
    'planes_trimesh',
    'local_matrix',
    'local_matrix_external',
    'assembly_Cf',
    'A_heights',
    'A_stress',
    'hessian',
    'simple_nurbs'
]


def planes_trimesh(form):

    # Planes Of the Triangles

    for key in form.faces():
        vs = form.face_vertices(key)
        p1 = array(form.vertex_coordinates(vs[0]))
        p2 = array(form.vertex_coordinates(vs[1]))
        p3 = array(form.vertex_coordinates(vs[2]))
        # print(p1,p2,p3)
        v1 = p3 - p1
        v2 = p2 - p1
        cp = cross(v1, v2)
        a, b, c = cp
        # d = -1 * dot(cp, p3)
        # sol = [a, b, c, d]
        form.set_face_attribute(key, 'dx', value=a/abs(a))
        form.set_face_attribute(key, 'dy', value=b/abs(b))

        # print('Sol:')
        # print(sol)

    return


def local_matrix(form, key, plot=False):

    k_i = form.key_index()
    uv_i = form.uv_index()

    if plot:
        print('Internal -> key: {0} , index: {1}'.format(key, k_i[key]))

    neighbors = form.vertex_neighbors(key, ordered=True)

    if is_ccw_xy(form.vertex_coordinates(key), form.vertex_coordinates(neighbors[0]), form.vertex_coordinates(neighbors[1])) is True:
        pass
    else:
        neighbors.reverse()

    if plot:
        print('neighbors:', neighbors)
    i = 0

    hn = []
    kn = []
    ln = []

    uv_lg = {}
    k_lg = {}

    for v in neighbors:
        edge = array(form.vertex_coordinates(v)[:2] + [0]) - array(form.vertex_coordinates(key)[:2] + [0])
        hi = cross(edge, [0, 0, 1])
        hn.append(normalize_vector(hi)[:2])
        ln.append(normrow(edge))
        kn.append(normalize_vector(edge)[:2])
        try:
            i_edge = uv_i[(v, key)]
        except Exception:
            i_edge = uv_i[(key, v)]
        uv_lg[i] = i_edge
        k_lg[i] = k_i[v]
        i += 1

    k_lg[i] = k_i[key]

    pn = len(neighbors)

    # local matrix

    j_ = zeros((pn, pn+1), dtype=int8)
    j__ = zeros((pn, pn+1), dtype=int8)
    j___ = zeros((pn, pn+1), dtype=int8)

    for j in range(pn):
        for k in range(pn+1):
            j_[j][k] = j
            if j_[j][k] > 0:
                j__[j][k] = j_[j][k] - 1
            else:
                j__[j][k] = pn - 1
            if j_[j][k] < pn - 1:
                j___[j][k] = j_[j][k] + 1
            else:
                j___[j][k] = 0

    # print(j_)
    # print(j__)
    # print(j___)

    C = zeros((pn, pn+1))

    for j in range(pn):
        for k in range(pn+1):
            m = j_[j][k]
            n = j__[j][k]
            p = j___[j][k]
            C[j][k] = 0.0
            if k == m:
                C[j][k] = dot(hn[p], hn[m])/(ln[m] * dot(hn[p], kn[m])) - dot(hn[n], hn[m])/(ln[m] * dot(hn[n], kn[m]))
            if k == n:
                C[j][k] = -1 * dot(hn[m], hn[m])/(ln[n] * dot(hn[m], kn[n]))
            if k == p:
                C[j][k] = dot(hn[m], hn[m])/(ln[p] * dot(hn[m], kn[p]))
            if k == pn:
                C[j][k] = -1 * (dot(hn[p], hn[m])/(ln[m] * dot(hn[p], kn[m])) - dot(hn[n], hn[m])/(ln[m] * dot(hn[n], kn[m]))) - 1 * \
                    (-1 * dot(hn[m], hn[m])/(ln[n] * dot(hn[m], kn[n]))) - 1 * (dot(hn[m], hn[m])/(ln[p] * dot(hn[m], kn[p])))

    # print('Local Matrix C int')
    # print(C.shape)
    # print(k_lg)
    # print(uv_lg)

    return C, k_lg, uv_lg


def local_matrix_external(form, key, plot=False):

    k_i = form.key_index()
    uv_i = form.uv_index()

    neighbors = form.vertex_neighbors(key, ordered=True)
    if is_ccw_xy(form.vertex_coordinates(key), form.vertex_coordinates(neighbors[0]), form.vertex_coordinates(neighbors[1])) is True:
        pass
    else:
        neighbors.reverse()
        if plot:
            print('Reversed Local Matrix')

    boundary = form.vertices_on_boundary()

    if plot:
        print('External -> key: {0} , index: {1}'.format(key, k_i[key]))

    # Shift list so it starts in the one in boundary

    if neighbors[0] not in boundary:
        if plot:
            print('Shift')
        for i in range(1, len(neighbors)):
            if neighbors[i] in boundary:
                neighbors = neighbors[i:] + neighbors[:i]

    if plot:
        print('neighbors:', neighbors)
    i = 0
    j = 0

    hn = []
    kn = []
    ln = []

    uv_lg = {}
    k_lg = {}

    for v in neighbors:
        edge = array(form.vertex_coordinates(v)[:2] + [0]) - array(form.vertex_coordinates(key)[:2] + [0])
        hi = cross(edge, [0, 0, 1])
        hn.append(normalize_vector(hi)[:2])
        ln.append(normrow(edge))
        kn.append(normalize_vector(edge)[:2])
        try:
            i_edge = uv_i[(v, key)]
        except BaseException:
            i_edge = uv_i[(key, v)]
        if form.is_edge_on_boundary(key, v) is False:
            uv_lg[j] = i_edge
            j += 1
        k_lg[i] = k_i[v]
        i += 1

    k_lg[i] = k_i[key]

    # quantity of neighbors in boundary
    nj = len(neighbors[1:len(neighbors)-1])  # in reality
    pn = len(neighbors)

    # local matrix

    j_ = zeros((nj, pn+1), dtype=int8)
    j__ = zeros((nj, pn+1), dtype=int8)
    j___ = zeros((nj, pn+1), dtype=int8)

    for j in range(nj):
        for k in range(pn+1):
            j_[j][k] = j + 1
            j__[j][k] = j + 2
            j___[j][k] = j + 2

    # print(j_)
    # print(j__)
    # print(j___)

    C = zeros((nj, pn+1))

    for j in range(nj):
        for k in range(pn+1):
            m = j_[j][k]
            n = j__[j][k]
            p = j___[j][k]
            C[j][k] = 0.0
            if k == m:
                C[j][k] = dot(hn[p], hn[m])/(ln[m] * dot(hn[p], kn[m])) - dot(hn[n], hn[m])/(ln[m] * dot(hn[n], kn[m]))
            if k == n:
                C[j][k] = -1 * dot(hn[m], hn[m])/(ln[n] * dot(hn[m], kn[n]))
            if k == p:
                C[j][k] = dot(hn[m], hn[m])/(ln[p] * dot(hn[m], kn[p]))
            if k == pn:
                C[j][k] = -1 * (dot(hn[p], hn[m])/(ln[m] * dot(hn[p], kn[m])) - dot(hn[n], hn[m])/(ln[m] * dot(hn[n], kn[m]))) - 1 * \
                    (-1 * dot(hn[m], hn[m])/(ln[n] * dot(hn[m], kn[n]))) - 1 * (dot(hn[m], hn[m])/(ln[p] * dot(hn[m], kn[p])))

    # print('Local Matrix C ext')
    # print(C.shape)
    # print(k_lg)
    # print(uv_lg)

    return C, k_lg, uv_lg


def assembly_Cf(form, plot=False):

    Cn = {}
    nodes_lg = {}
    edges_lg = {}

    k_i = form.key_index()
    uv_i = form.uv_index()

    for key in form.vertices():
        i = k_i[key]
        if key not in form.vertices_on_boundary():
            Cn[i], nodes_lg[i], edges_lg[i] = local_matrix(form, key, plot=plot)
        else:
            Cn[i], nodes_lg[i], edges_lg[i] = local_matrix_external(form, key, plot=plot)

    # Build Global Matrix Cf

    N = form.number_of_vertices()
    E = form.number_of_edges()
    ext_edges = []
    for u, v in form.edges_on_boundary():
        try:
            ext_edges.append(uv_i[(u, v)])
        except BaseException:
            ext_edges.append(uv_i[(v, u)])
    edges_int = E - len(form.edges_on_boundary())
    print('Form has {0} interior edges\n   {1} bound edges\n   {2} total nodes\n   {3} boundaries nodes'.format(edges_int, len(ext_edges), N, len(form.vertices_on_boundary())))
    Cf = zeros((E, N))
    if plot:
        print(Cf.shape)
    for i in range(N):
        Clocal = Cn[i]
        node_lg = nodes_lg[i]
        edge_lg = edges_lg[i]
        for j in range(Clocal.shape[0]):
            for k in range(Clocal.shape[1]):
                j_ = edge_lg[j]
                k_ = node_lg[k]
                Cf[j_][k_] += Clocal[j][k]

    for u, v in form.edges_on_boundary():
        try:
            i = uv_i[(u, v)]
        except BaseException:
            i = uv_i[(v, u)]

    return Cf


def A_heights(form):

    uv_i = form.uv_index()
    k_i = form.key_index()

    N = form.number_of_vertices()
    E = form.number_of_edges()
    A = zeros((N, E))

    for key in form.vertices():
        ngb = form.vertex_neighbors(key, ordered=True)
        zi = form.vertex_attribute(key, 'z')
        i_vertex = k_i[key]
        for m in ngb:
            if m not in form.edges_on_boundary():
                zm = form.vertex_attribute(m, 'z')
                try:
                    i_edge = uv_i[key, m]
                    h_ij = form.edge_length(key, m)
                except BaseException:
                    i_edge = uv_i[m, key]
                    h_ij = form.edge_length(m, key)
                val = (zi - zm)/h_ij
                A[i_vertex][i_edge] += val

    return A


def A_stress(form, Rm):

    uv_i = form.uv_index()
    k_i = form.key_index()

    N = form.number_of_vertices()
    A = zeros((N, N))

    for key in form.vertices():
        ngb = form.vertex_neighbors(key, ordered=True)
        i_hor = k_i[key]
        for m in ngb:
            if m not in form.edges_on_boundary():
                i_ver = k_i[m]
                try:
                    i_edge = uv_i[key, m]
                    h_ij = form.edge_length(key, m)
                except BaseException:
                    i_edge = uv_i[m, key]
                    h_ij = form.edge_length(m, key)
                Rmij = Rm[i_edge]
                val = Rmij/h_ij
                A[i_hor][i_ver] += - 1 * val
                A[i_hor][i_hor] += val

    return A


def hessian(form):

    # Method per Node

    for key in form.vertices():
        area = form.vertex_attribute(key, 'pz')
        neighbors = form.vertex_neighborhood(key)
        jump = array([[0, 0], [0, 0]])
        u = key
        for v in neighbors:
            edge = array(form.vertex_coordinates(v)[:2] + [0]) - array(form.vertex_coordinates(u)[:2] + [0])
            hi = cross(edge, [0, 0, 1])
            hi_ = normalize_vector(hi)
            Hi = tensorproduct(hi_, hi_)
            # print(Hi) # Check if it should be planar
            df = form.vertex_coordinates(v)[2] - form.vertex_coordinates(u)[2]
            jump += df * Hi
        form.vertex_attribute(key, 'jump', jump)
        form.vertex_attribute(key, 'hessian', jump/area)
        print(jump/area)

    # Method per edge connected to the node

    for key in form.vertices():
        area = form.vertex_attribute(key, 'pz')
        neighbors = form.vertex_neighborhood(key)
        hess = array([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
        u = key
        for v in neighbors:
            try:
                dfHi = form.edge_attribute((u, v), 'jump')
            except BaseException:
                dfHi = form.edge_attribute((v, u), 'jump')
            hess += dfHi
        form.vertex_attribute(key, 'hessian', hess)
        print(hess)

    return


def simple_nurbs(a, b, h, par=20, plot=False):

    from geomdl import CPGen
    from geomdl import BSpline
    # from geomdl import NURBS
    from geomdl import utilities
    from geomdl.visualization import VisMPL
    from matplotlib import cm

    # Generate a plane with the dimensions 50x100
    surfgrid = CPGen.Grid(a, b)

    # Generate a grid of 25x30
    surfgrid.generate(2, 2)

    # Generate bumps on the grid
    # surfgrid.bumps(num_bumps=5, bump_height=20, base_extent=8)

    # Create a BSpline surface instance
    surf = BSpline.Surface()

    # Set degrees
    surf.degree_u = 2
    surf.degree_v = 2

    # Get the control points from the generated grid
    surf.ctrlpts2d = surfgrid.grid
    surf.ctrlpts2d[1][1][2] = h

    # Set knot vectors
    surf.knotvector_u = utilities.generate_knot_vector(surf.degree_u, surf.ctrlpts_size_u)
    surf.knotvector_v = utilities.generate_knot_vector(surf.degree_v, surf.ctrlpts_size_v)
    # print(surf.knotvector_u)

    # Set sample size
    surf.sample_size = par

    if plot:
        # Set visualization component
        surf.vis = VisMPL.VisSurface(ctrlpts=True, legend=False)

        # Plot the surface
        surf.render(colormap=cm.get_cmap('cool'))

    return surf




from compas.datastructures import mesh_smooth_centroid
from compas.datastructures import mesh_smooth_area
from compas.datastructures import mesh_smooth_centerofmass

from compas.geometry import closest_point_on_line

from compas_tno.algorithms.equilibrium import equilibrium_fdm


def constrained_smoothing(mesh, kmax=100, damping=0.5,  constraints={}, algorithm='centroid'):
    """Constrained smoothing of a mesh. Constraints can be points and lines.

    Parameters
    ----------
    mesh : :class:`~compas.datastructures.Mesh`
        A mesh to smooth.
    kmax : int
        Number of iterations for smoothing. Default value 100.
    damping : float
        Damping value for smoothing between 0 and 1. Default value 0.5.
    constraints : dict
        Dictionary of constraints as vertex keys pointing to fixed points or lines.
    algorithm : string
        Type of smoothing algorithm to apply (classic centroid or area-based). Classic centroid by default.

    Return
    ----------
    mesh: :class:`~compas.datastructures.Mesh`
        The smoothed mesh.

    Notes
    ----------
    This function was extracted from `compas singular <https://blockresearchgroup.github.io/compas_singular/>`_ developed by Robin Oval.

    """

    def callback(k, args):

        mesh, constraints = args

        for vkey, constraint in constraints.items():
            if constraint is None:
                continue
            elif len(constraint) == 3:  # It is a Point
                x, y, z = constraint
            elif len(constraint) == 2:  # It is a Line
                x, y, z = closest_point_on_line(mesh.vertex_coordinates(vkey), constraint)
            else:
                continue

            mesh.vertex[vkey]['x'] = x
            mesh.vertex[vkey]['y'] = y
            mesh.vertex[vkey]['z'] = z

    func = {'centroid': mesh_smooth_centroid, 'area': mesh_smooth_area, 'centerofmass': mesh_smooth_centerofmass}

    if algorithm not in func:
        algorithm = 'centroid'

    func[algorithm](mesh, kmax=kmax, damping=damping, callback=callback, callback_args=[mesh, constraints])

    return mesh


def apply_sag(form, boundary_force=10.0, signe_compression=-1.0):  # probably move location
    """Relax the mesh with FDM assuming higher force in the boudary elements.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The formdiagram to apply sag.
    boundary_force : float
        Force density in the edges on the boundary.
        The default value is ``10.0``.
    signe_compression : float
        Indicate the default sign for compression.
        The default value is ``-1.0``.

    Return
    ----------
    form: :class:`~compas_tno.diagrams.FormDiagram`
        The relaxed form diagram.

    """

    for u, v in form.edges():
        form.edge_attribute((u, v), 'q', signe_compression*1.0)

    for u, v in form.edges_on_boundary():
        form.edge_attribute((u, v), 'q', signe_compression*boundary_force)

    equilibrium_fdm(form)

    for key in form.vertices():
        form.vertex_attribute(key, 'z', 0.0)

    return form



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas_tno.rhino.diagramobject import DiagramObject


__all__ = ['ForceObject']


class ForceObject(DiagramObject):
    """A force object represents a force diagram in the Rhino view.
    """

    SETTINGS = {
        'show.vertices': True,
        'show.edges': True,
        'show.vertexlabels': False,
        'show.edgelabels': False,
        'show.forcelabels': False,
        'show.forcecolors': True,

        'color.vertices': (0, 0, 0),
        'color.vertexlabels': (255, 255, 255),
        'color.vertices:is_fixed': (0, 255, 255),
        'color.edges': (0, 0, 0),
        'color.edges:is_ind': (0, 255, 255),
        'color.edges:is_external': (0, 255, 0),
        'color.edges:is_reaction': (0, 255, 0),
        'color.edges:is_load': (0, 255, 0),
        'color.faces': (210, 210, 210),
        'color.compression': (0, 0, 255),
        'color.tension': (255, 0, 0),

        'tol.forces': 1e-3,
    }

    def __init__(self, diagram, *args, **kwargs):
        super(ForceObject, self).__init__(diagram, *args, **kwargs)
        self.settings.update(ForceObject.SETTINGS)
        settings = kwargs.get('settings') or {}
        if settings:
            self.settings.update(settings)

    def draw(self):
        """Draw the diagram.

        The visible components, display properties and visual style of the diagram
        can be fully customised using the configuration items in the settings dict.

        The method will clear the drawing layer and any objects it has drawn in a previous call,
        and keep track of any newly created objects using their GUID.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.clear()
        if not self.visible:
            return

        self.artist.vertex_xyz = self.vertex_xyz

        # vertices
        if self.settings['show.vertices']:
            vertices = list(self.diagram.vertices())
            color = {}
            color.update({vertex: self.settings['color.vertices'] for vertex in vertices})
            color.update({vertex: self.settings['color.vertices:is_fixed'] for vertex in self.diagram.vertices_where({'is_fixed': True})})
            guids = self.artist.draw_vertices(color=color)
            self.guid_vertex = zip(guids, vertices)

            # vertex labels
            if self.settings['show.vertexlabels']:
                text = {vertex: index for index, vertex in enumerate(vertices)}
                color = {}
                color.update({vertex: self.settings['color.vertexlabels'] for vertex in vertices})
                color.update({vertex: self.settings['color.vertices:is_fixed'] for vertex in self.diagram.vertices_where({'is_fixed': True})})
                guids = self.artist.draw_vertexlabels(text=text, color=color)
                self.guid_vertexlabel = zip(guids, vertices)

        # edges
        if self.settings['show.edges']:
            tol = self.settings['tol.forces']
            edges = [edge for edge in self.diagram.edges() if self.diagram.edge_length(*edge) > tol]
            color = {}
            color.update({edge: self.settings['color.edges'] for edge in edges})
            color.update({edge: self.settings['color.edges:is_external'] for edge in self.diagram.edges_where_dual({'is_external': True})})
            color.update({edge: self.settings['color.edges:is_load'] for edge in self.diagram.edges_where_dual({'is_load': True})})
            color.update({edge: self.settings['color.edges:is_reaction'] for edge in self.diagram.edges_where_dual({'is_reaction': True})})
            color.update({edge: self.settings['color.edges:is_ind'] for edge in self.diagram.edges_where_dual({'is_ind': True})})

            # force colors
            if self.settings['show.forcecolors']:
                tol = self.settings['tol.forces']
                for edge in self.diagram.edges_where_dual({'is_external': False}):
                    if self.diagram.dual_edge_force(edge) > + tol:
                        color[edge] = self.settings['color.tension']
                    elif self.diagram.dual_edge_force(edge) < - tol:
                        color[edge] = self.settings['color.compression']

            guids = self.artist.draw_edges(edges=edges, color=color)
            self.guid_edge = zip(guids, edges)

            guid_edgelabel = []

            # edge labels
            if self.settings['show.edgelabels']:
                edge_index = self.diagram.edge_index(self.diagram.dual)
                edge_index.update({(v, u): index for (u, v), index in edge_index.items()})
                text = {edge: edge_index[edge] for edge in edges}
                color = {}
                color.update({edge: self.settings['color.edges'] for edge in edges})
                color.update({edge: self.settings['color.edges:is_external'] for edge in self.diagram.edges_where_dual({'is_external': True})})
                color.update({edge: self.settings['color.edges:is_load'] for edge in self.diagram.edges_where_dual({'is_load': True})})
                color.update({edge: self.settings['color.edges:is_reaction'] for edge in self.diagram.edges_where_dual({'is_reaction': True})})
                color.update({edge: self.settings['color.edges:is_ind'] for edge in self.diagram.edges_where_dual({'is_ind': True})})

                # force colors
                if self.settings['show.forcecolors']:
                    tol = self.settings['tol.forces']
                    for edge in self.diagram.edges_where_dual({'is_external': False}):
                        if self.diagram.dual_edge_force(edge) > + tol:
                            color[edge] = self.settings['color.tension']
                        elif self.diagram.dual_edge_force(edge) < - tol:
                            color[edge] = self.settings['color.compression']

                guids = self.artist.draw_edgelabels(text=text, color=color)
                guid_edgelabel += zip(guids, edges)

            # force labels
            if self.settings['show.forcelabels']:
                text = {}
                for edge in edges:
                    f = self.diagram.dual_edge_force(edge)
                    text[edge] = "{:.4g}kN".format(abs(f))

                color = {}
                color.update({edge: self.settings['color.edges'] for edge in edges})
                color.update({edge: self.settings['color.edges:is_external'] for edge in self.diagram.edges_where_dual({'is_external': True})})
                color.update({edge: self.settings['color.edges:is_load'] for edge in self.diagram.edges_where_dual({'is_load': True})})
                color.update({edge: self.settings['color.edges:is_reaction'] for edge in self.diagram.edges_where_dual({'is_reaction': True})})
                color.update({edge: self.settings['color.edges:is_ind'] for edge in self.diagram.edges_where_dual({'is_ind': True})})

                # force colors
                if self.settings['show.forcecolors']:
                    tol = self.settings['tol.forces']
                    for edge in self.diagram.edges_where_dual({'is_external': False}):
                        if self.diagram.dual_edge_force(edge) > + tol:
                            color[edge] = self.settings['color.tension']
                        elif self.diagram.dual_edge_force(edge) < - tol:
                            color[edge] = self.settings['color.compression']

                guids = self.artist.draw_edgelabels(text=text, color=color)

                guid_edgelabel += zip(guids, edges)

            self.guid_edgelabel = guid_edgelabel

        self.redraw()



from compas.geometry import distance_point_point_xy
from compas.geometry import distance_point_line_xy
from compas.geometry import closest_point_on_line_xy

from numpy import zeros

import matplotlib.pyplot as plt


__all__ = [
    'apply_radial_symmetry',
    'apply_symmetry_from_axis',
    'find_sym_axis_in_rect_patterns',
    'build_symmetry_matrix',
    'build_symmetry_transformation',
    'build_vertex_symmetry_transformation',
    'build_symmetry_matrix_supports'
]


def apply_radial_symmetry(form, center=[5.0, 5.0, 0.0], correct_loads=True):
    """Apply a radial symmetry based on a center points. Applicable for dome circulat patterns.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram of the problem
    center : [float, float, float], optional
        The coordinates of the center of the pattern, by default [5.0, 5.0, 0.0]
    correct_loads : bool, optional
        Whether or not the loads should be corrected in the nodes for perfect symmetry, by default True

    Returns
    -------
    None
        The FormDiagram is modified in place.
    """

    form.edges_attribute('sym_key', None)
    form.vertices_attribute('sym_key', None)

    dist_checked = []
    dist_dict = {}

    # Symmetry on the independent edges

    for u, v in form.edges_where({'_is_edge': True}):
        midpoint = form.edge_midpoint(u, v)
        dist = round(distance_point_point_xy(center, midpoint), 10)
        dist_dict[(u, v)] = dist
        if dist not in dist_checked:
            dist_checked.append(dist)
    i = 0
    for dist in dist_checked:
        for u, v in dist_dict:
            if dist_dict[(u, v)] == dist:
                form.edge_attribute((u, v), 'sym_key', i)

        i += 1

    # Symmetry on the Support's position

    dist_checked = []
    dist_dict = {}

    for key in form.vertices():
        point = form.vertex_coordinates(key)
        dist = round(distance_point_point_xy(center, point), 10)
        dist_dict[key] = dist
        if dist not in dist_checked:
            dist_checked.append(dist)
    i = 0
    for dist in dist_checked:
        pass_first = False
        pz = None
        ub = None
        lb = None
        for key in dist_dict:
            if dist_dict[key] == dist:
                form.vertex_attribute(key, 'sym_key', i)
                if correct_loads:
                    if not pass_first:
                        pass_first = True
                        pz = form.vertex_attribute(key, 'pz')
                        ub = form.vertex_attribute(key, 'ub')
                        lb = form.vertex_attribute(key, 'lb')
                        s = form.vertex_attribute(key, 'target')
                    form.vertex_attribute(key, 'pz', pz)
                    form.vertex_attribute(key, 'ub', ub)
                    form.vertex_attribute(key, 'lb', lb)
                    form.vertex_attribute(key, 'target', s)
        i += 1

    return


def apply_symmetry_from_axis(form, list_axis_symmetry=[], correct_loads=True, tol=0.01):
    """ Apply a symmetry based on a series of axis of symmetry. Applicable for rectangular patterns.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram of the problem
    list_axis_symmetry : list, optional
        The list of the axis of symmetry, by default []
    correct_loads : bool, optional
        Whether or not the loads should be corrected in the nodes for perfect symmetry, by default True
    tol : float, optional
        Tollerance to assume symmetry, by default 0.001

    Returns
    -------
    None
        The FormDiagram is modified in place.

    """

    form.edges_attribute('sym_dict', None)
    form.vertices_attribute('sym_dict', None)
    form.edges_attribute('sym_key', None)
    form.vertices_attribute('sym_key', None)

    i = 0
    for axis_symmetry in list_axis_symmetry:
        axis_str = str(axis_symmetry)
        dist_checked = []
        dist_dict = {}
        for u, v in form.edges_where({'_is_edge': True}):
            midpoint = form.edge_midpoint(u, v)
            dist_line = round(distance_point_line_xy(midpoint, axis_symmetry), 10)  # change this logic to something that takes into account tol
            closest_pt = closest_point_on_line_xy(midpoint, axis_symmetry)  # geometric_key(..) used before
            # dist = [closest_pt, dist_line]
            dist = [dist_line, closest_pt]
            dist_dict[(u, v)] = dist
            if dist not in dist_checked:
                dist_checked.append(dist)
        for dist in dist_checked:
            for u, v in dist_dict:
                # if dist_dict[(u, v)] == dist:
                if abs(dist[0] - dist_dict[(u, v)][0]) < tol and distance_point_point_xy(dist[1], dist_dict[(u, v)][1]) < tol:
                    form.edge_attribute((u, v), 'sym_key', i)
                    if not form.edge_attribute((u, v), 'sym_dict'):
                        dic = {}
                    else:
                        dic = form.edge_attribute((u, v), 'sym_dict')
                    dic[axis_str] = i
                    form.edge_attribute((u, v), 'sym_dict', dic)
            i += 1

    if len(list_axis_symmetry) > 1:
        checked_edge = []
        groups_of_edges = []
        for edge in form.edges_where({'_is_edge': True}):
            if edge in checked_edge:
                continue
            group = [edge]
            checked_edge.append(edge)
            values = list(form.edge_attribute(edge, 'sym_dict').values())
            for edge2 in form.edges_where({'_is_edge': True}):
                if edge2 in checked_edge:
                    continue
                values2 = list(form.edge_attribute(edge2, 'sym_dict').values())
                if len(set(values2) & set(values)):
                    group.append(edge2)
                    checked_edge.append(edge2)
                    values = list(set(values + values2))
            groups_of_edges.append(group)

        i = 0
        for group in groups_of_edges:
            form.edges_attribute('sym_key', value=i, keys=group)
            i += 1

    i = 0
    for axis_symmetry in list_axis_symmetry:
        axis_str = str(axis_symmetry)
        dist_checked = []
        dist_dict = {}
        for key in form.vertices():
            point = form.vertex_coordinates(key)
            dist_line = round(distance_point_line_xy(point, axis_symmetry), 10)
            closest_pt = closest_point_on_line_xy(point, axis_symmetry)  # geometric_key( .. )
            # dist = [closest_pt, dist_line]
            dist = [dist_line, closest_pt]
            dist_dict[key] = dist
            if dist not in dist_checked:
                dist_checked.append(dist)
        for dist in dist_checked:
            for key in dist_dict:
                # if dist_dict[key] == dist:
                if abs(dist[0] - dist_dict[key][0]) < tol and distance_point_point_xy(dist[1], dist_dict[key][1]) < tol:
                    form.vertex_attribute(key, 'sym_key', i)
                    if not form.vertex_attribute(key, 'sym_dict'):
                        dic = {}
                    else:
                        dic = form.vertex_attribute(key, 'sym_dict')
                    dic[axis_str] = i
                    form.vertex_attribute(key, 'sym_dict', dic)
            i += 1

    if len(list_axis_symmetry) > 1:
        checked_vertex = []
        groups_of_vertices = []
        for vertex in form.vertices():
            if vertex in checked_vertex:
                continue
            group = [vertex]
            checked_vertex.append(vertex)
            values = list(form.vertex_attribute(vertex, 'sym_dict').values())
            for vertex2 in form.vertices():
                if vertex2 in checked_vertex:
                    continue
                values2 = list(form.vertex_attribute(vertex2, 'sym_dict').values())
                if len(set(values2) & set(values)):
                    group.append(vertex2)
                    checked_vertex.append(vertex2)
                    values = list(set(values + values2))
            groups_of_vertices.append(group)

        i = 0
        for group in groups_of_vertices:
            form.vertices_attribute('sym_key', value=i, keys=group)
            i += 1
            if correct_loads:
                pz = form.vertex_attribute(group[0], 'pz')
                ub = form.vertex_attribute(group[0], 'ub')
                lb = form.vertex_attribute(group[0], 'lb')
                s = form.vertex_attribute(group[0], 'target')
                form.vertices_attribute('pz', value=pz, keys=group)
                form.vertices_attribute('ub', value=ub, keys=group)
                form.vertices_attribute('lb', value=lb, keys=group)
                form.vertices_attribute('target', value=s, keys=group)

    return


def find_sym_axis_in_rect_patterns(data_form):
    """ Find the axis of symmetry in rectangular patterns.

    Parameters
    ----------
    data_form : dict
        A dictionary with the parameters used to create the FormDiagram.

    Returns
    -------
    lines: list
        A list of the symmetry lines.

    """

    lines = []
    xy_span = data_form.get('xy_span', None)
    if xy_span:
        [[x0, x1], [y0, y1]] = xy_span
        xc = (x0 + x1)/2
        yc = (y0 + y1)/2
        hor_line = [[x0, yc], [x1, yc]]
        ver_line = [[xc, y0], [xc, y1]]
        diag_line = [[x0, y0], [x1, y1]]
        lines = [hor_line, ver_line, diag_line]
    else:
        raise NotImplementedError

    return lines


def build_symmetry_matrix(form, printout=False):
    """ Build a symmetry matrix such as Asym * q = 0, with Asym shape (m - k; m)

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram of the problem
    printout : bool, optional
        Whether or not display messages are printed, by default True

    Returns
    -------
    Asym: array (m - k x m)
        The symmetry matrix.
    """

    m = len(list(form.edges_where({'_is_edge': True})))
    k_unique = form.number_of_sym_edges(printout=printout)
    Asym = zeros((m - k_unique, m))
    uv_i = form.uv_index()

    line = 0
    for id_sym in range(k_unique):
        i = 0
        for u, v in form.edges():
            if form.edge_attribute((u, v), 'sym_key') == id_sym:
                index = uv_i[(u, v)]
                if i == 0:
                    index0 = index
                else:
                    Asym[line, index0] = 1
                    Asym[line, index] = -1
                    line += 1
                i += 1
    if printout:
        plt.matshow(Asym)
        plt.colorbar()
        plt.show()

    return Asym


def build_symmetry_transformation(form, printout=False):
    """Build a symmetry matrix Esym (m, k) such as q = Esym * qsym.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram of the problem
    printout : bool, optional
        Whether or not display messages are printed, by default True

    Returns
    -------
    Esym: array (m x k)
        The symmetry matrix.

    """

    m = len(list(form.edges_where({'_is_edge': True})))
    k_unique = form.number_of_sym_edges(printout=printout)
    Esym = zeros((m, k_unique))
    uv_i = form.uv_index()

    for id_sym in range(k_unique):
        Ei = zeros((m, 1))
        for u, v in form.edges():
            if form.edge_attribute((u, v), 'sym_key') == id_sym:
                index = uv_i[(u, v)]
                Ei[index] = 1.0
        Esym[:, id_sym] = Ei.flatten()

    if printout:
        plt.matshow(Esym)
        plt.colorbar()
        plt.show()

    return Esym


def build_vertex_symmetry_transformation(form, printout=False):
    """Build a symmetry matrix Evsym (n, k) such as z = Evsym * zb.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram of the problem
    printout : bool, optional
        Whether or not display messages are printed, by default True

    Returns
    -------
    Evsym: array (n x k)
        The symmetry matrix.
    """

    n = form.number_of_edges()
    k_unique = form.number_of_sym_vertices(printout=printout)
    Evsym = zeros((n, k_unique))
    k_i = form.key_index()

    for id_sym in range(k_unique):
        Ei = zeros((n, 1))
        for key in form.vertices():
            if form.vertex_attribute(key, 'sym_key') == id_sym:
                index = k_i[key]
                Ei[index] = 1.0
        Evsym[:, id_sym] = Ei.flatten()

    if printout:
        plt.matshow(Evsym)
        plt.colorbar()
        plt.show()

    return Evsym


def build_symmetry_matrix_supports(form, printout=False):
    """Build a symmetry matrix to the supports.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram of the problem
    printout : bool, optional
        Whether or not display messages are printed, by default True

    Returns
    -------
    Evsym: array (n x k)
        The symmetry matrix.
    """

    n = form.number_of_supports()
    k_unique = form.number_of_sym_supports(printout=printout)
    Asym = zeros((n - k_unique, n))
    key_i_sup = {}

    i = 0
    for key in form.vertices_where({'is_fixed': True}):
        key_i_sup[key] = i
        i += 1

    line = 0
    for id_sym in range(k_unique):
        i = 0
        for key in form.vertices_where({'is_fixed': True}):
            if form.vertex_attribute(key, 'sym_key') == id_sym:
                index = key_i_sup[key]
                if i == 0:
                    index0 = index
                else:
                    Asym[line, index0] = 1
                    Asym[line, index] = -1
                    line += 1
                i += 1
    if printout:
        plt.matshow(Asym)
        plt.colorbar()
        plt.show()

    return Asym



from compas_tno.solvers.solver_MATLAB import run_loadpath_from_form_MATLAB
from compas_tno.solvers.solver_cvxpy import run_loadpath_from_form_CVXPY

from compas_tno.algorithms import form_update_with_parallelisation
from compas_tno.algorithms import equilibrium_fdm
from compas_tno.algorithms import compute_reactions

import importlib.util


def initialize_loadpath(form, problem=None, find_inds=False, solver_convex='CVXPY', printout=False):
    """Built-in function to optimise the loadpath considering diagram fixed projection.
    Note: This function will select the most appropriate solver (CVX or MOSEK)

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram to compute the loadpath
    problem : :class:`~compas_tno.problems.Problem`, optional
        The problem classs with the matrices relevant of the problem, by default None
    find_inds : bool, optional
        If independents need to be found before the loadpath computation, by default False
    solver_convex : str, optional
        Solver to compute the convex optimisation, by default CVXPY
    printout : bool, optional
        If prints about the optimisation data should appear in the screen, by default False

    Returns
    -------
    problem : Problem
        The class with the main matrices of the problem
    """

    if solver_convex == 'CVX' or solver_convex == 'MATLAB':
        if not importlib.util.find_spec('matlab'):
            raise ValueError('MATLAB/CVX not configured. Try changing the <solver_convex> attribute.')
        problem = run_loadpath_from_form_MATLAB(form, problem=problem, find_inds=find_inds, printout=printout)
    elif solver_convex == 'CVXPY' or solver_convex == 'MOSEK':
        if not importlib.util.find_spec('cvxpy'):
            raise ValueError('CVXPY/MOSEK not configured. Try changing the <solver_convex> attribute.')
        problem = run_loadpath_from_form_CVXPY(form, problem=problem, find_inds=find_inds, printout=printout)
    else:
        raise ValueError('Could not initilalise loadpath optimisation with {}. Try changing the <solver_convex> attribute.'.format(solver_convex))

    return problem


def initialize_tna(form, plot=False):
    """Initialize the equilibrium in a form diagram with applied loads using TNA interative solver procedure (form and force diagrams are parallel)

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram. Loads and support must already have been assigned
    plot : bool, optional
        Plots of the intermediare and final force diagrams to follow the process, by default False
    """

    form_update_with_parallelisation(form, plot=plot)

    compute_reactions(form)


def initialize_fdm(form):
    """Initialize the equilibrium in a form diagram with applied loads using FDM approach for the q's stored in the form

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram. Loads, supports and force densities must already have been assigned
    """

    equilibrium_fdm(form)

    compute_reactions(form)



import math
from numpy import array


__all__ = [
    'check_envelope_constraints',
    'distance_target',
    'rectangular_smoothing_constraints',
    'assign_cracks',
    'rollers_on_openings',
    'set_b_constraint',
    'set_rollers_constraint'
]


def check_envelope_constraints(form, tol=1e-6):
    """Check if the envelope constraints are respected and return a ``penalty > 0`` if not.

    Parameters
    ----------
    form : FormDiagram
        The form diagram with the envelope as nodal attributes.
    tol : float, optional
        Tolerance for verifying the constraints, by default ``1e-6``.

    Returns
    -------
    penalty: float
        The penalty computed as the sum square of the exceeding value at each node.
    """
    penalty = 0.0

    for key in form.vertices():
        _, _, z = form.vertex_coordinates(key)
        lb = form.vertex_attribute(key, 'lb')
        ub = form.vertex_attribute(key, 'ub')
        if z < lb - tol:
            outside = (lb - z)
            penalty += outside**2
        if z > ub + tol:
            outside = (z - ub)
            penalty += outside**2

    print('The penalty in the constraints is {0:.3f}'.format(penalty))

    return penalty


def distance_target(form):
    """Returns a measure of the distance among the TN and the target surface.

    Parameters
    ----------
    form : FormDiagram
        The form diagram to check.

    Returns
    -------
    distance: float
        The square distance is computed.

    """

    dist = 0

    for key in form.vertices():
        _, _, z = form.vertex_coordinates(key)
        targ = form.vertex_attribute(key, 'target')
        dist += (z - targ) ** 2

    return dist


def rectangular_smoothing_constraints(form, xy_span=[[0, 10], [0, 10]]):
    """Constraint for smoothing the form diagram in a rectangular boundary.

    Parameters
    ----------
    form : FormDiagram
        The form diagram to check.
    xy_span: list, optional
        The rectangular footprint to smooth, by default [[0, 10], [0, 10]]

    Returns
    -------
    distance: float
        The square distance is computed.

    """

    [[x0, x1], [y0, y1]] = xy_span
    cons = {key: None for key in form.vertices()}
    line_top = [[x0, y1, 0], [x1, y1, 0]]
    line_bottom = [[x0, y0, 0], [x1, y0, 0]]
    line_left = [[x0, y0, 0], [x0, y1, 0]]
    line_right = [[x1, y0, 0], [x1, y1, 0]]
    for key in form.vertices_on_boundary():
        x, y, z = form.vertex_coordinates(key)
        if x == x0:
            cons[key] = line_left
        elif x == x1:
            cons[key] = line_right
        elif y == y0:
            cons[key] = line_bottom
        elif y == y1:
            cons[key] = line_top
    for key in form.vertices_where({'is_fixed': True}):
        cons[key] = form.vertex_coordinates(key)
    return cons


def assign_cracks(form, dx=[[0.50, 0.55]], dy=[[-0.1, 0.1]], type=['top']):
    """Assign cracks on a form diagram to the nodes desired

    Parameters
    ----------
    form : FormDiagram
        ForceDiagram to constraint
    dx : list, optional
        List of the range on the x-coordinates of the nodes to be constrained, by default [[0.50, 0.55]]
    dy : list, optional
        List of the range on the y-coordinates of the nodes to be constrained, by default [[-0.1, 0.1]]
    type : list, optional
        List with the type of constraint to applye to the vertices (top or bottom), by default ['top']

    Returns
    -------
    form : FormDiagram
        ForceDiagram with the constraints in attribute 'cracks'.
    """

    k_i = form.key_index()
    cracks_ub = []
    cracks_lb = []

    # count = 0

    for key in form.vertices():
        x, y, _ = form.vertex_coordinates(key)
        for i in range(len(dx)):
            if dx[i][0] <= x <= dx[i][1] and dy[i][0] <= y <= dy[i][1]:
                if type[i] == 'top':
                    cracks_ub.append(k_i[key])
                if type[i] == 'bottom':
                    cracks_lb.append(k_i[key])

    form.attributes['cracks'] = (cracks_lb, cracks_ub)

    return form


def rollers_on_openings(form, xy_span=[[0.0, 10.0], [0.0, 10.0]], max_f=5.0, constraint_directions='all'):
    """ Apply rollers to the rectangular boundaries of the pattern.

    Parameters
    ----------
    form : FormDiagram
        ForceDiagram to constraint.
    xy_span: list, optional
        The rectangular footprint to smooth, by default [[0, 10], [0, 10]]
    max_f : float, optional
        The maximum force allowed in each roller, by default 5.0
    constraint_directions : str
        Define the restriction in ``all`` directions or only ``x``or only ``y``.

    Returns
    -------
    form : FormDiagram
        ForceDiagram with the rollers assigned.
    """

    y1 = xy_span[1][1]
    y0 = xy_span[1][0]
    x1 = xy_span[0][1]
    x0 = xy_span[0][0]

    bndr = form.vertices_on_boundary()

    for key in bndr:
        if form.vertex_attribute(key, 'is_fixed') is False:
            x, y, _ = form.vertex_coordinates(key)
            if x == x1 and (constraint_directions in ['all', 'x']):
                form.vertex_attribute(key, 'rol_x', True)
                form.vertex_attribute(key, 'max_rx', max_f)
            if x == x0 and (constraint_directions in ['all', 'x']):
                form.vertex_attribute(key, 'rol_x', True)
                form.vertex_attribute(key, 'max_rx', max_f)
            if y == y1 and (constraint_directions in ['all', 'y']):
                form.vertex_attribute(key, 'rol_y', True)
                form.vertex_attribute(key, 'max_ry', max_f)
            if y == y0 and (constraint_directions in ['all', 'y']):
                form.vertex_attribute(key, 'rol_y', True)
                form.vertex_attribute(key, 'max_ry', max_f)

    return form


def circular_joints(form, x0=None, xf=None, blocks=18, thk=0.5, t=0.0, tol=1e-3):
    """Deprecated function to assign constraints in the joints among blocks instead of in the z's.

    Parameters
    ----------
    form : FormDiagram
        The form diagram
    x0 : float, optional
        The initial ordinate of the arch, by default None
    xf : float, optional
        The final ordinate of the arch, by default None
    blocks : int, optional
        Number of blocks, by default 18
    thk : float, optional
        The thickness of the arch, by default 0.5
    t : float, optional
        Lower bound in the vertices of the boundary, by default 0.0
    tol : float, optional
        The tolerance, by default 1e-3

    Returns
    -------
    form : FormDiagram
        The form diagram with the joints
    """

    k_i = form.key_index()

    if x0 is None or xf is None:
        x = []
        for key in form.vertices():
            x.append(form.vertex_coordinates(key)[0])
        x0 = min(x)
        xf = max(x)
    y = 0.0

    xc = (xf+x0)/2
    r = xf - xc
    ri = r - thk/2
    re = r + thk/2
    form.attributes['Re'] = re
    form.attributes['Ri'] = ri
    print('SpanMid: {0:.2} m / SpanInt: {1:.2} m / SpanExt: {2:.2} m / Thickness: {3:.4} m / Ratio t/Ri: {4:.4} m / Ratio t/R: {5:.4} m / Number of Blocks: {6}'.format(2 *
          r, 2*ri, 2*re, thk, (thk/ri), (thk/r), blocks))

    njoints = blocks+1
    joints = {}
    for j in range(njoints):
        theta = j/blocks*math.pi
        xi = xc + ri * math.cos(theta)  # takeout
        zi = ri * math.sin(theta) + tol
        xe = xc + re * math.cos(theta)
        ze = re * math.sin(theta) + tol  # take out
        xmax = max(xi, xe)
        xmin = min(xi, xe)
        possible_edges = []
        for (u, v) in form.edges():
            xu, xv = form.vertex_coordinates(u)[0], form.vertex_coordinates(v)[0]
            if max(xu, xv) >= xmin and min(xu, xv) <= xmax:
                possible_edges.append(tuple(sorted([k_i[u], k_i[v]])))
                if form.vertex_attribute(u, 'is_fixed'):
                    possible_edges.append(tuple(sorted([-k_i[u], k_i[u]])))
                if form.vertex_attribute(v, 'is_fixed'):
                    possible_edges.append(tuple(sorted([-k_i[v], k_i[v]])))
        joints[j] = [[xi, y, zi], [xe, y, ze], set(possible_edges)]
        print(joints[j])
    form.attributes['joints'] = joints

    for key in form.vertices():
        x, _, _ = form.vertex_coordinates(key)
        zt = math.sqrt(r**2 - (x-xc)**2)
        ze = math.sqrt(re**2 - (x-xc)**2) - t
        form.vertex_attribute(key, 'target', value=zt)
        zi2 = ri**2 - (x-xc)**2
        if zi2 < 0:
            zi = 0 - t
        else:
            zi = math.sqrt(zi2) - t
        if form.vertex_attribute(key, 'is_fixed'):
            form.vertex_attribute(key, 'lb', value=None)
            form.vertex_attribute(key, 'ub', value=None)
        else:
            form.vertex_attribute(key, 'lb', value=zi)
            form.vertex_attribute(key, 'ub', value=ze)
        # form.vertex_attribute(key,'z',value=ze)
        if form.vertex_attribute(key, 'is_fixed'):
            form.vertex_attribute(key, 'b', value=[thk/2, 0.0])
        if x == x0:
            form.attributes['tmax'] = ze

    return form


def set_b_constraint(form, printout=False):
    """Helper to set the ``b`` constraint on the reactions of the structure.

    Parameters
    ----------
    form : FormDiagram
        The form diagram to apply the constraints
    printout : bool, optional
        If prints are added to thr screen, by default False

    Returns
    -------
    array (nb x 2)
        The limits for the reaction force.
    """
    b = []
    for key in form.vertices_where({'is_fixed': True}):
        try:
            [b_] = form.vertex_attributes(key, 'b')
            b.append(b_)
        except BaseException:
            pass
    b = array(b)
    if printout:
        print('Reaction bounds active in : {0} joints'.format(len(b)))
    return b


def set_rollers_constraint(form, printout=False):
    """Helper to constraints on rollers.

    Parameters
    ----------
    form : FormDiagram
        The form diagram to apply the constraints
    printout : bool, optional
        If prints are added to thr screen, by default False

    Returns
    -------
    array (nb x 2)
        The limits for the reaction force.
    """
    max_rol_rx = []
    max_rol_ry = []
    for key in form.vertices_where({'rol_x': True}):
        max_rol_rx.append(form.vertex_attribute(key, 'max_rx'))
    for key in form.vertices_where({'rol_y': True}):
        max_rol_ry.append(form.vertex_attribute(key, 'max_ry'))
    if printout:
        print('Constraints on rollers activated in {0} in x and {1} in y.'.format(len(max_rol_rx), len(max_rol_ry)))
    return array(max_rol_rx).reshape(-1, 1), array(max_rol_ry).reshape(-1, 1)



from compas_tno.shapes import Shape
from compas_tno.diagrams import FormDiagram
from compas_tno.analysis import Analysis
from compas_tno.viewers import Viewer
from compas_tno.plotters import TNOPlotter

# ----------------------------------------
# 1. Geometric definition
# ----------------------------------------
H = 1.0
L = 2.0
b = 0.5
discretisation = 20

arch = Shape.create_arch(H=H, L=L, b=b)

# ----------------------------------------
# 2. Form Diagram
# ----------------------------------------

form = FormDiagram.create_arch(H=H, L=L, discretisation=discretisation)

# ----------------------------------------
# 3. Create analysis for minimum thrust
# ----------------------------------------
analysis = Analysis.create_minthrust_analysis(form, arch)
analysis.optimiser.set_constraints(['funicular', 'envelope', 'reac_bounds'])
analysis.apply_selfweight()
analysis.apply_envelope()
analysis.apply_reaction_bounds()
analysis.set_up_optimiser()
analysis.run()

# ----------------------------------------
# 4. Visualise solution
# ----------------------------------------

view = Viewer(form, arch)
view.settings['camera.target'] = [0.5, 0, 0]
view.settings['camera.distance'] = 7.0
view.settings['camera.rx'] = 60
view.draw_form()
view.draw_shape()
view.draw_cracks()
view.show()

# ----------------------------------------
# 5. Create analysis for maximum thrust and visualise
# ----------------------------------------

analysis = Analysis.create_maxthrust_analysis(form, arch)
analysis.optimiser.set_constraints(['funicular', 'envelope', 'reac_bounds'])
analysis.apply_selfweight()
analysis.apply_envelope()
analysis.apply_reaction_bounds()
analysis.set_up_optimiser()
analysis.run()

view = Viewer(form, arch)
view.settings['camera.target'] = [0.5, 0, 0]
view.settings['camera.distance'] = 7.0
view.settings['camera.rx'] = 60
view.draw_form()
view.draw_shape()
view.draw_cracks()
view.draw_reactions(extend_reactions=True)
view.show()



"""
********************************************************************************
compas_tno.solvers
********************************************************************************

.. currentmodule:: compas_tno.solvers

Scipy
========

.. autosummary::
    :toctree: generated/

    run_optimisation_scipy

IPOPT
========

.. autosummary::
    :toctree: generated/

    run_optimisation_ipopt
    Wrapper_ipopt
    Wrapper_ipopt_autodiff

MATLAB
========

.. autosummary::
    :toctree: generated/

    run_optimisation_MATLAB
    run_loadpath_from_form_MATLAB
    call_and_output_CVX_MATLAB

CVXPY
========

.. autosummary::
    :toctree: generated/

    run_optimisation_CVXPY
    run_loadpath_from_form_CVXPY
    call_and_output_CVXPY

PyOpt
========

.. autosummary::
    :toctree: generated/

    run_optimisation_pyOpt

MMA
========

.. autosummary::
    :toctree: generated/

    run_optimisation_MMA
    mma_numpy

Post Processing
===============

.. autosummary::
    :toctree: generated/

    post_process_general

"""
from __future__ import absolute_import

from .mma_numpy import mma_numpy
from .solver_scipy import (
    run_optimisation_scipy
)
from .solver_IPOPT import (
    run_optimisation_ipopt,
    Wrapper_ipopt,
    Wrapper_ipopt_autodiff
)
from .solver_MATLAB import (
    run_optimisation_MATLAB,
    run_loadpath_from_form_MATLAB,
    call_and_output_CVX_MATLAB,
)
from .solver_cvxpy import (
    run_optimisation_CVXPY,
    run_loadpath_from_form_CVXPY,
    call_and_output_CVXPY
)
from .post_process import post_process_general
from .solver_MMA import run_optimisation_MMA
from .solver_pyOpt import run_optimisation_pyOpt


__all__ = [
    'run_optimisation_scipy',

    'run_optimisation_ipopt',
    'Wrapper_ipopt',
    'Wrapper_ipopt_autodiff',

    'mma_numpy',
    'run_optimisation_pyOpt',

    'run_optimisation_MATLAB',
    'run_loadpath_from_form_MATLAB',
    'call_and_output_CVX_MATLAB',

    'run_optimisation_CVXPY',
    'run_loadpath_from_form_CVXPY',
    'call_and_output_CVXPY',

    'post_process_general',

    'run_optimisation_MMA',

]



import json
import compas_tno
from numpy import array

from compas_tno.algorithms import xyz_from_xopt
from compas_tno.algorithms import reciprocal_from_form


def callback_save_json(xopt, *args, **kwargs):
    """Save the variables in the ``output.json`` file created.

    Parameters
    ----------
    xopt : array
        The variables in one iteration of the optimisation
    """

    DATA_FILENAME = compas_tno.get('output.json')

    with open(DATA_FILENAME, mode='r', encoding='utf-8') as f:
        data = json.load(f)

    i = len(data['iterations'])
    data['iterations'][i] = list(xopt.flatten())

    with open(DATA_FILENAME, mode='w', encoding='utf-8') as f:
        json.dump(data, f)

    return


def callback_create_json():
    """Create a ``output.json`` to store the iterations of an optimisation"""

    data = {'iterations': {}}

    DATA_FILENAME = compas_tno.get('output.json')

    with open(DATA_FILENAME, mode='w', encoding='utf-8') as f:
        json.dump(data, f)

    return


def save_geometry_at_iterations(form, optimiser, force=False):
    """Save the geometry of the form (and force) during iterations of the optimisation. Works only with SLSQP and IPOPT solvers.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The Form Diagram of the problem
    optimiser : :class:`~compas_tno.optimisers.Optimiser`
        The Optimiser with the numerical information about the problem
    force : bool, optional
        Whether or not the force diagram coordinates should also be saved, by default False

    Notes
    -----
        Xforce
            Address where the geometry of the structure at each step is saved ``Xform.json``
        Xforce
            Address where the geometry of the force diagram at each step is saves ``Xforce.json``.

    """

    M = optimiser.M  # matrices of the problem

    file_qs = compas_tno.get('output.json')
    file_Xform = compas_tno.get('Xform.json')
    file_Xforce = None

    if force:
        file_Xforce = compas_tno.get('Xforce.json')
        Xforce = {}

    with open(file_qs, mode='r', encoding='utf-8') as f:
        data = json.load(f)

    Xform = {}

    iterations = len(data['iterations'])

    for i in range(iterations):
        xopt_i = array(data['iterations'][str(i)]).reshape(-1, 1)
        M = xyz_from_xopt(xopt_i, M)
        Xform_i = M.X.tolist()
        Xform[str(i)] = Xform_i

        if force:
            j = 0
            for key in form.vertices():
                form.vertex_attributes(key, 'xyz', M.X[j].tolist())
                j += 1
            k = 0
            for edge in form.edges_where({'_is_edge': True}):
                form.edge_attribute(edge, 'q', M.q.flatten()[k])
                k += 1

            force = reciprocal_from_form(form)
            Xforce_i = force.vertices_attributes('xyz')
            Xforce[str(i)] = Xforce_i

    with open(file_Xform, mode='w', encoding='utf-8') as f:
        json.dump(Xform, f)

    print('Form Geometry saved @:', file_Xform)

    if force:
        with open(file_Xforce, mode='w', encoding='utf-8') as f:
            json.dump(Xforce, f)

        print('Force Geometry saved @:', file_Xforce)

    return file_Xform, file_Xforce




from .post_process import post_process_general
from compas_tno.algorithms import xyz_from_q
from compas_tno.algorithms import q_from_qid
from numpy import hstack


__all__ = [
    'run_optimisation_pyOpt'
]


def run_optimisation_pyOpt(analysis):
    """Run Optimisation using PyOpt (requires installation).

    Parameters
    ----------
    analysis : :class:`~compas_tno.analysis.Analysis`
        The Analysis object to pass.

    Returns
    -------
    None
        Analysis is executed in place.

    Note
    -------
        Requires pyOpt installation.
    """

    import pyOpt

    form = analysis.form
    optimiser = analysis.optimiser
    solver = optimiser.settings['solver']
    objective = optimiser.settings['objective']
    variables = optimiser.settings['variables']
    fobj = optimiser.fobj
    fconstr = optimiser.fconstr
    args = optimiser.args
    x0 = optimiser.x0
    g0 = optimiser.g0
    f0 = optimiser.fo
    (q, ind, dep, E, Edinv, Ei, C, Ct, Ci, Cit, Cf, U, V, p, px, py, pz, z, free, fixed, lh, sym, k, lb, ub, lb_ind, ub_ind, s, Wfree, x, y, b, joints,
     cracks_lb, cracks_ub, free_x, free_y, rol_x, rol_y, Citx, City, Cftx, Cfty) = args
    i_uv = form.index_uv()
    k_i = form.key_index()
    bounds = optimiser.bounds
    x0 = optimiser.x0

    if solver.split('-')[0] == 'pyOpt':
        lower = [lw[0] for lw in bounds]
        upper = [up[1] for up in bounds]
        solver_pyOpt = solver.split('-')[1]
        title = 'Solver: ' + solver_pyOpt + 'Objective: ' + objective

        opt_prob = pyOpt.Optimization(title, pyOpt_wrapper)
        opt_prob.addObj('f', value=f0)
        opt_prob.addVarGroup('x', len(x0), type='c', value=x0, lower=lower, upper=upper)
        opt_prob.addConGroup('c', len(g0), type='i', value=g0)

        if solver_pyOpt == 'SLSQP':
            slv = pyOpt.SLSQP()
            slv.setOption('MAXIT', 1000)

        if solver_pyOpt == 'PSQP':
            slv = pyOpt.PSQP()

        if solver_pyOpt == 'CONMIN':
            slv = pyOpt.CONMIN()

        if solver_pyOpt == 'COBYLA':
            slv = pyOpt.COBYLA()

        if solver_pyOpt == 'SOLVOPT':
            slv = pyOpt.SOLVOPT()

        if solver_pyOpt == 'KSOPT':
            slv = pyOpt.KSOPT()

        if solver_pyOpt == 'NSGA2':
            slv = pyOpt.NSGA2()

        if solver_pyOpt == 'ALGENCAN':
            slv = pyOpt.ALGENCAN()

        if solver_pyOpt == 'FILTERSD':
            slv = pyOpt.FILTERSD()

        if solver_pyOpt == 'SDPEN':
            slv = pyOpt.SDPEN()

        if solver_pyOpt == 'ALPSO':
            slv = pyOpt.SDPEN()

        if solver_pyOpt == 'ALHSO':
            slv = pyOpt.SDPEN()

        if solver_pyOpt == 'MIDACO':
            slv = pyOpt.SDPEN()

        [fopt, xopt, info] = slv(opt_prob, sens_type='FD', args=args, objective=fobj, constraints=fconstr)
        # exitflag = info['text']
        if info['text'] == 'Optimization terminated successfully.':
            exitflag = 0
        else:
            exitflag = 1
        # exitflag = slv.informs
        # if printout:
        #     print(opt_prob.solution(0))
        print(info['text'])
        fopt = fopt.item()
        niter = 100
        if 'zb' in variables:
            q[ind] = xopt[:k].reshape(-1, 1)
            z[fixed] = xopt[k:].reshape(-1, 1)
        else:
            q[ind] = xopt[:k].reshape(-1, 1)

    args = (q, ind, dep, E, Edinv, Ei, C, Ct, Ci, Cit, Cf, U, V, p, px, py, pz, z, free, fixed, lh, sym, k, lb, ub, lb_ind, ub_ind, s, Wfree, x, y, b, joints, i_uv, k_i)
    q = q_from_qid(q, ind, Edinv, Ei, p)
    Pi = hstack([px, py, pz])
    Xb = hstack([x[fixed], y[fixed], z[fixed]])
    X = xyz_from_q(q, Pi, Xb, Ci, Cit)
    z = X[:, [2]]

    optimiser.niter = niter
    optimiser.exitflag = exitflag
    optimiser.fopt = fopt

    post_process_general(analysis)

    return analysis


def pyOpt_wrapper(x, **kwargs):
    args = kwargs['args']
    fobj = kwargs['objective']
    fconst = kwargs['constraints']
    # Temp
    f = fobj(x, *args)
    g = -1 * fconst(x, *args)  # pyOpt get constraints g1 <= 0 as default
    fail = 0
    return f, g, fail



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas_rhino
from compas_tno.rhino.diagramobject import DiagramObject
from compas.geometry import add_vectors
from compas.geometry import scale_vector
from compas.geometry import norm_vector
from compas.geometry import centroid_points


__all__ = ['FormObject']


class FormObject(DiagramObject):
    """A form object represents a form diagram in the Rhino model space.
    """

    SETTINGS = {
        'show.vertices': False,
        'show.supports': True,
        'show.edges': True,
        'show.vertexlabels': False,
        'show.vertexloads': False,
        'show.edgelabels': False,
        'show.reactionlabels': False,
        'show.cracks': False,
        'show.vertices_bound': False,
        'show.forcecolors': False,
        'show.forcelabels': False,
        'show.forcepipes': False,
        'show.loadvectors': False,
        'show.reactionvectors': False,

        'is.compression_negative': True,

        'color.vertices': (0, 0, 0),
        'color.vertexlabels': (255, 255, 255),
        'color.vertexloads': (255, 255, 255),
        'color.vertices:is_fixed': (255, 0, 0),
        'color.vertices:upper_bound': (0, 200, 0),
        'color.vertices:lower_bound': (0, 0, 200),
        'color.edges': (0, 0, 0),
        'color.edges:is_ind': (0, 0, 0),
        'color.edges:is_external': (0, 255, 0),
        'color.edges:is_reaction': (0, 255, 0),
        'color.edges:is_load': (0, 255, 0),
        'color.reactionvectors': (255, 0, 0),
        'color.loadvectors': (0, 200, 200),
        'color.faces': (210, 210, 210),
        'color.compression': (255, 0, 0),
        'color.tension': (0, 0, 255),

        'scale.forcepipes': 0.001,
        'scale.vectors': 0.01,

        'tol.edges': 1e-3,
        'tol.forces': 1e-3,
    }

    def __init__(self, diagram, *args, **kwargs):
        super(FormObject, self).__init__(diagram, *args, **kwargs)
        self.settings.update(FormObject.SETTINGS)
        settings = kwargs.get('settings') or {}
        if settings:
            self.settings.update(settings)
        self._guid_pipes = {}
        self._guid_cracks = {}
        self._guid_vectors = {}
        self._guid_reactions = {}

    @property
    def guids(self):
        guids = super(FormObject, self).guids
        guids += list(self.guid_pipes.keys())
        guids += list(self.guid_cracks.keys())
        guids += list(self.guid_vectors.keys())
        guids += list(self.guid_reactions.keys())
        return guids

    @property
    def guid_pipes(self):
        """Map between Rhino object GUIDs and form diagram edge force identifiers."""
        return self._guid_pipes

    @guid_pipes.setter
    def guid_pipes(self, values):
        self._guid_pipes = dict(values)

    @property
    def guid_cracks(self):
        """Map between Rhino object GUIDs and thrust network cracks."""
        return self._guid_cracks

    @guid_cracks.setter
    def guid_cracks(self, values):
        self._guid_cracks = dict(values)

    @property
    def guid_vectors(self):
        """Map between Rhino object GUIDs and thrust network vectors to represent loads."""
        return self._guid_vectors

    @guid_vectors.setter
    def guid_vectors(self, values):
        self._guid_vectors = dict(values)

    @property
    def guid_reactions(self):
        """Map between Rhino object GUIDs and thrust network vectors to represent reactions."""
        return self._guid_reactions

    @guid_reactions.setter
    def guid_reactions(self, values):
        self._guid_reactions = dict(values)

    def clear(self):
        super(FormObject, self).clear()
        compas_rhino.delete_objects(self.guids, purge=True)
        self._guid_pipes = {}
        self._guid_cracks = {}
        self._guid_vectors = {}
        self._guid_reactions = {}

    def draw(self):
        """Draw the form diagram.

        The visible components, display properties and visual style of the form diagram
        drawn by this method can be fully customised using the configuration items
        in the settings dict: ``FormArtist.settings``.

        The method will clear the scene of any objects it has previously drawn
        and keep track of any newly created objects using their GUID.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.clear()
        if not self.visible:
            return

        self.artist.vertex_xyz = self.vertex_xyz

        guids = []
        vertices = []

        # vertices (including supports)
        if self.settings['show.vertices']:
            vertices = list(self.diagram.vertices())
            color = {}
            color.update({vertex: self.settings['color.vertices'] for vertex in vertices})
            color.update({vertex: self.settings['color.vertices:is_fixed'] for vertex in self.diagram.vertices_where({'is_fixed': True})})
            guids = self.artist.draw_vertices(color=color)
            self.guid_vertex = zip(guids, vertices)

            # vertex labels
            if self.settings['show.vertexlabels']:
                text = {vertex: index for index, vertex in enumerate(vertices)}
                color = {}
                color.update({vertex: self.settings['color.vertexlabels'] for vertex in vertices})
                color.update({vertex: self.settings['color.vertices:is_fixed'] for vertex in self.diagram.vertices_where({'is_fixed': True})})
                guids = self.artist.draw_vertexlabels(text=text, color=color)
                self.guid_vertexlabel = zip(guids, vertices)

        # supports only
        elif self.settings['show.supports']:
            vertices = list(self.diagram.vertices_where({'is_fixed': True}))
            color = {}
            color.update({vertex: self.settings['color.vertices:is_fixed'] for vertex in self.diagram.vertices_where({'is_fixed': True})})
            guids = self.artist.draw_vertices(vertices=vertices, color=color)
            self.guid_vertex = zip(guids, vertices)

        # upper and lower bounds
        if self.settings['show.vertices_bound']:
            points = []
            vertices = []
            for key in self.diagram.vertices():
                lb = self.diagram.vertex_attribute(key, 'lb')
                ub = self.diagram.vertex_attribute(key, 'ub')
                x, y, z = self.diagram.vertex_coordinates(key)
                if lb:
                    vertices.append(key)
                    points.append({
                        'pos': [x, y, lb],
                        'color': self.settings['color.vertices:lower_bound']
                    })
                if ub:
                    vertices.append(key)
                    points.append({
                        'pos': [x, y, ub],
                        'color': self.settings['color.vertices:upper_bound']
                    })

            guids = compas_rhino.draw_points(points, layer="TNO::Shape::Bounds", clear=False, redraw=False)
            self.guid_cracks = zip(guids, vertices)

        if self.settings['show.cracks']:
            guids, vertices = self.artist.draw_cracks(
                color_intrados=self.settings['color.vertices:lower_bound'],
                color_extrados=self.settings['color.vertices:upper_bound'],
                layer='TNO::Shape'
            )
            self.guid_cracks = zip(guids, vertices)

        # vertex loads
        if self.settings['show.vertexloads']:
            vertices = list(self.diagram.vertices())
            text = {vertex: '{0:.1f}'.format(self.diagram.vertex_attribute(vertex, 'pz')) for vertex in self.diagram.vertices()}
            color = {}
            color.update({vertex: self.settings['color.vertexloads'] for vertex in vertices})
            color.update({vertex: self.settings['color.vertices:is_fixed'] for vertex in self.diagram.vertices_where({'is_fixed': True})})
            guids = self.artist.draw_vertexlabels(text=text, color=color)
            self.guid_vertexlabel = zip(guids, vertices)

        # edges
        if self.settings['show.edges']:
            edges = list(self.diagram.edges_where({'_is_edge': True}))
            color = {}
            color.update({edge: self.settings['color.edges'] for edge in edges})
            color.update({edge: self.settings['color.edges:is_external'] for edge in self.diagram.edges_where({'is_external': True})})
            color.update({edge: self.settings['color.edges:is_load'] for edge in self.diagram.edges_where({'is_load': True})})
            color.update({edge: self.settings['color.edges:is_reaction'] for edge in self.diagram.edges_where({'is_reaction': True})})
            color.update({edge: self.settings['color.edges:is_ind'] for edge in self.diagram.edges_where({'is_ind': True})})

            # force colors
            if self.settings['show.forcecolors']:
                tol = self.settings['tol.forces']
                for edge in self.diagram.edges_where({'is_external': False}):
                    if self.diagram.edge_attribute(edge, 'f') > + tol:
                        color[edge] = self.settings['color.tension']
                    elif self.diagram.edge_attribute(edge, 'f') < - tol:
                        color[edge] = self.settings['color.compression']

            guids = self.artist.draw_edges(edges=edges, color=color)
            self.guid_edge = zip(guids, edges)

            guid_edgelabel = []

            # edge labels
            if self.settings['show.edgelabels']:
                text = {edge: index for index, edge in enumerate(edges)}
                color = {}
                color.update({edge: self.settings['color.edges'] for edge in edges})
                color.update({edge: self.settings['color.edges:is_external'] for edge in self.diagram.edges_where({'is_external': True})})
                color.update({edge: self.settings['color.edges:is_load'] for edge in self.diagram.edges_where({'is_load': True})})
                color.update({edge: self.settings['color.edges:is_reaction'] for edge in self.diagram.edges_where({'is_reaction': True})})
                color.update({edge: self.settings['color.edges:is_ind'] for edge in self.diagram.edges_where({'is_ind': True})})

                # force colors
                if self.settings['show.forcecolors']:
                    tol = self.settings['tol.forces']
                    for edge in self.diagram.edges_where({'is_external': False}):
                        if self.diagram.edge_attribute(edge, 'f') > + tol:
                            color[edge] = self.settings['color.tension']
                        elif self.diagram.edge_attribute(edge, 'f') < - tol:
                            color[edge] = self.settings['color.compression']

                guids = self.artist.draw_edgelabels(text=text, color=color)
                guid_edgelabel += zip(guids, edges)

            # force labels
            if self.settings['show.forcelabels']:
                text = {}
                for index, edge in enumerate(self.diagram.edges_where({'is_external': True})):
                    f = self.diagram.edge_attribute(edge, 'f')
                    if f != 0.0:
                        text[edge] = "{:.4g}kN".format(abs(f))
                color = {}
                # color.update({edge: self.settings['color.edges'] for edge in self.diagram.edges()})
                color.update({edge: self.settings['color.edges:is_external'] for edge in self.diagram.edges_where({'is_external': True})})
                color.update({edge: self.settings['color.edges:is_load'] for edge in self.diagram.edges_where({'is_load': True})})
                color.update({edge: self.settings['color.edges:is_reaction'] for edge in self.diagram.edges_where({'is_reaction': True})})
                color.update({edge: self.settings['color.edges:is_ind'] for edge in self.diagram.edges_where({'is_ind': True})})

                guids = self.artist.draw_edgelabels(text=text, color=color)
                guid_edgelabel += zip(guids, edges)

            # reaction labels
            if self.settings['show.reactionlabels']:
                vertices = list(self.diagram.vertices_where({'is_fixed': True}))
                labels = []
                for vertex in vertices:
                    a = self.vertex_xyz[vertex]
                    r = self.diagram.vertex_attributes(vertex, ['_rx', '_ry', '_rz'])
                    if not any(r):  # If receives null vector or None
                        text = "0.0 kN"
                        b = a
                    else:
                        text = "{:.4g}kN".format(norm_vector(r))
                        r = scale_vector(r, -1 * self.settings['scale.vectors'])
                        b = add_vectors(a, r)
                    labels.append({
                        'pos': centroid_points([a, b]),
                        'name': "{}.reactionlabel.{}".format(self.diagram.name, vertex),
                        'color': self.settings['color.reactionvectors'],
                        'text': text})

                guids = compas_rhino.draw_labels(labels, layer=self.layer, clear=False, redraw=False)
                guid_edgelabel += zip(guids, vertices)

            self.guid_edgelabel = guid_edgelabel

        # force pipes
        if self.settings['show.forcepipes']:
            edges = list(self.diagram.edges_where({'_is_edge': True}))
            guids = self.artist.draw_forcepipes(
                color_compression=self.settings['color.compression'],
                color_tension=self.settings['color.tension'],
                compression_negative=self.settings['is.compression_negative'],
                scale=self.settings['scale.forcepipes'],
                tol=self.settings['tol.forces'])

            self.guid_pipes = zip(guids, edges)

        # load vectors
        if self.settings['show.loadvectors']:
            vertices = list(self.diagram.vertices())
            guids = self.artist.draw_loads(
                color=self.settings['color.loadvectors'],
                scale=self.settings['scale.vectors'],
                layer="TNO::Loads",
                tol=self.settings['tol.forces']
            )

            self.guid_vectors = zip(guids, edges)

        # reaction vectors
        if self.settings['show.reactionvectors']:
            vertices = list(self.diagram.vertices_where({'is_fixed': True}))
            guids = self.artist.draw_reactions(
                color=self.settings['color.reactionvectors'],
                scale=self.settings['scale.vectors'],
                layer="TNO::Reactions",
                tol=self.settings['tol.forces']
            )

            self.guid_reactions = zip(guids, edges)

        self.redraw()



def rectangular_topology(x, y):
    """Draw the vertices position and dictionary for a rectangular base besh

    Parameters
    ----------
    x : list
        List with the x-coordinates required for the mesh
    y : list
        List with the y-coordinates required for the mesh

    Returns
    -------
    xi: list
        List with the x-coordinate of the vertices.
    yi: list
        List with the y-coordinate of the vertices.
    faces_i: list of lists
        Lists with the connectivity required to create the mesh.
    """

    index = 0
    uv_i = {}
    faces = []
    faces_i = []

    xi = []
    yi = []

    for i in range(len(x)):
        for j in range(len(y)):
            uv_i[(i, j)] = index
            xi.append(x[i])
            yi.append(y[j])

            if i < len(x) - 1 and j < len(y) - 1:
                p1 = (i, j)
                p2 = (i, j+1)
                p3 = (i+1, j)
                p4 = (i+1, j+1)
                if i != j and i + j != len(x) - 2:
                    face = [p1, p2, p4, p3]
                    faces.append(face)
                else:
                    if i == j:
                        faces.append([p1, p2, p4])
                        faces.append([p1, p4, p3])
                    else:
                        faces.append([p2, p3, p1])
                        faces.append([p2, p4, p3])
            index = index + 1

    for face in faces:
        face_i = []
        for uv in face:
            u, v = uv
            i = uv_i[(u, v)]
            face_i.append(i)
        faces_i.append(face_i)

    return xi, yi, faces_i




from numpy import array
# from numpy import max
# from numpy import min
from numpy import newaxis
from numpy import zeros

from scipy.sparse.linalg import spsolve
from scipy.sparse import diags

from compas.numerical import connectivity_matrix

from compas_tno.algorithms import equilibrium_fdm
from compas_tno.algorithms import vertical_equilibrium_fdm

# from compas_tno.diagrams.form import evaluate_a
# from compas_tno.diagrams.form import energy
# from compas_tno.diagrams.form import loadpath

# from compas_tna.diagrams import ForceDiagram

from compas_tno.plotters import plot_form
# from compas_tno.plotters import plot_force
# from compas_tno.plotters import plot_grad

from copy import deepcopy


__all__ = [
    'optimise_fdm',
    'optimise_tna',
]


def optimise_fdm(form, plot=True, surf=False):

    # Mapping

    k_i = form.key_index()

    # Vertices and edges

    vertices = [k_i[key] for key in form.vertices_where({'_is_external': False})]
    edges = [[k_i[u], k_i[v]] for u, v in form.edges_where({'_is_edge': True, '_is_external': False})]
    fixed = [k_i[key] for key in form.vertices_where({'_is_external': False, 'is_fixed': True})]
    free = list(set(vertices) - set(fixed))

    n = len(vertices)

    # Co-ordinates and loads

    xyz = zeros((n, 3))
    x = zeros((n, 1))
    y = zeros((n, 1))
    z = zeros((n, 1))
    s = zeros((n, 1))
    px = zeros((n, 1))
    py = zeros((n, 1))
    pz = zeros((n, 1))

    for key, vertex in form.vertex.items():
        if vertex.get('_is_external') is False:
            i = k_i[key]
            xyz[i, :] = form.vertex_coordinates(key)
            x[i] = vertex.get('x')
            y[i] = vertex.get('y')
            z[i] = vertex.get('z')
            s[i] = vertex.get('target')
            px[i] = vertex.get('px', 0)
            py[i] = vertex.get('py', 0)
            pz[i] = vertex.get('pz', 0)

    q = array([form.edge_attribute((u, v), 'q') for u, v in form.edges_where({'_is_edge': True, '_is_external': False})])[:, newaxis]

    # Interpolate Target Surface

    # if interpolate:
    #     from scipy import interpolate
    #     surf = interpolate.interp2d(x, y, s, kind = 'linear')

    # C and E matrices

    C = connectivity_matrix(edges, 'csr')
    Ci = C[:, free]
    Cit = Ci.transpose()

    # From Here calculate Iteractivelly q...

    it = 0
    it_max = 100
    stepsave = 10.0
    incrsave = 0.1

    while it < it_max:

        print('-'*10, ' Iteration {0}'.format(it))

        D = Cit.dot(diags(q.flatten())).dot(C)
        Di = D[:, free]

        # Grad Calculation

        # (L, values) = _chofactor(Di.toarray())

        W = diags(C.dot(z).flatten())  # dimension m x m
        J = spsolve(Di, Cit.dot(W))  # dimension ni x m

        # Y_ = spsolve(L,Cit.dot(W))
        # X_ = spsolve(L.transpose(),Y_)

        # Gradient of F

        # Grad = 2 * (diags(z[free,0].flatten()) - diags(s[free,0].flatten())).dot(X_)

        grad = (2 * (z[free] - s[free]).transpose() * J).transpose()

        step = stepsave
        incr = incrsave

        k = 0
        kmax = 300
        q0 = q
        f0 = form.loadpath()
        form_ = deepcopy(form)
        cont = True

        # Define steplength

        while cont:

            q = q0 + step * grad
            k += 1

            # form_ = update_form(form_, q)
            vertical_equilibrium_fdm(form_)
            f = form.loadpath()

            if f < f0:
                break
            else:
                if step/10 < incr:
                    incr = incr/2

                step = step - incr

                if k % 100 == 0:
                    print('Iteration steplength: {0} / Stepsize: {1} / Energy Var: {2:.1f}%'.format(k, step, (f-f0)/f0*100))

                if k > kmax:
                    print('Iterations on steplength reached the maximum {0}'.format(kmax))
                    break

        # Update Function

        if f < f0:
            form = form_
            stepsave = step
            incrsave = incr
            q = array([form.edge_attribute((u, v), 'q') for u, v in form.edges_where({'_is_edge': True})])[:, newaxis]
            for key, vertex in form_.vertex.items():
                i = k_i[key]
                z[i] = vertex.get('z')
            print('End of Iteration {0} / Stepsize: {1:.5f} / Evergy Var: {2:.1f}% / Energy: {3:.1f}'.format(it, step, (f-f0)/f0*100, f))

            form = equilibrium_fdm(form)

            if surf:
                for key, vertex in form.vertex.items():
                    xi, yi, _ = form.vertex_coordinates(key)
                    [si] = surf(xi, yi).tolist()
                    form.vertex_attribute(key, 'target', value=si)
                print('Updating constraints. Energy: {0:.1f}'.format(form.loadpath()))

        else:
            print('Optimisation not found - Iteration {0} / Stepsize: {1:.5f} / Energy: {2:.1f}'.format(it, step, f))
            break

        it += 1

        if plot:
            plot_form(form).show()

    return form


def optimise_tna(form, objective='target', plot=None, it_max=20, alpha=1.0, a_max=2.5, steplength=None, null=None, save_steps=False):

    # # Mapping

    # k_i = form.key_index()
    # uv_i = form.uv_index()

    # # Vertices and edges

    # vertices = [k_i[key] for key in form.vertices()]
    # edges = [[k_i[u], k_i[v]] for u, v in form.edges_where({'_is_edge': True})]
    # fixed = [k_i[key] for key in form.vertices_where({'is_anchor': True})]
    # free = list(set(vertices) - set(fixed))
    # if null:
    #     null_i = [uv_i[(u, v)] for u, v in null]

    # n = len(vertices)

    # # Co-ordinates and loads

    # xyz = zeros((n, 3))
    # x = zeros((n, 1))
    # y = zeros((n, 1))
    # z = zeros((n, 1))
    # s = zeros((n, 1))
    # px = zeros((n, 1))
    # py = zeros((n, 1))
    # pz = zeros((n, 1))
    # w = zeros((n, 1))

    # for key, vertex in form.vertex.items():
    #     i = k_i[key]
    #     xyz[i, :] = form.vertex_coordinates(key)
    #     x[i] = vertex.get('x')
    #     y[i] = vertex.get('y')
    #     z[i] = vertex.get('z')
    #     s[i] = vertex.get('target', 0)
    #     px[i] = vertex.get('px', 0)
    #     py[i] = vertex.get('py', 0)
    #     pz[i] = vertex.get('pz', 0)
    #     w[i] = vertex.get('weight', 1.0)

    # q = array([form.edge_attribute((u, v), 'q') for u, v in form.edges_where({'_is_edge': True})])[:, newaxis]
    # Wfree = diags(w[free].flatten())

    # C = connectivity_matrix(edges, 'csr')
    # Ci = C[:, free]
    # Cit = Ci.transpose()

    # if objective == 'target':
    #     grad_exp = grad_target
    #     f_exp = form.distance_target()
    #     print('Optimisation takes as objective minimise distance to target')
    # if objective == 'loadpath':
    #     grad_exp = grad_lp
    #     f_exp = form.loadpath()
    #     print('Optimisation takes as objective minimise loadpath')

    # args = (C, Ci, Cit, x, y, Wfree, s, free)

    # f = f_exp(form)
    # print('Objective Initial is f: {0}'.format(f))

    # it = 0
    # step = 100
    # print_ = 0
    # form.update_default_edge_attributes({'dq': 0.0})

    # while it < it_max:

    #     print('\n', '-'*20, '\n', ' Iteration {0}'.format(it))

    #     grad = grad_exp(q, z, args)

    #     print('grad range   : {0:.3f} : {1:.3f}'.format(min(grad), max(grad)))
    #     print('q start iteration range   : {0:.3f} : {1:.3f}'.format(min(q), max(q)))

    #     if steplength is not None:
    #         step = steplength
    #     else:
    #         for i in range(len(q)):
    #             if q[i] < 0:
    #                 print('FD something in tension')
    #             [t] = q[i]/grad[i]
    #             if t > 0:
    #                 if t < step:
    #                     step = t
    #     step = max([step, 0.1])
    #     incr = step/10

    #     print('Iterations start with steplength: {0}'.format(step))

    #     k = 0
    #     kmax = 200
    #     q0 = deepcopy(q)
    #     f0 = f_exp(form)
    #     print('Value of Energy - start: {0:.1f}'.format(f0))
    #     form_ = deepcopy(form)
    #     force_ = ForceDiagram.from_formdiagram(form_)
    #     cont = True
    #     a_limit = a_max

    #     print('-'*20)

    #     # Define steplength

    #     while cont:

    #         q = q0 - step * grad
    #         k += 1

    #         if null:
    #             q[null_i] *= 0

    #         if plot and k == 1:
    #             for u, v in form_.edges_where({'_is_edge': True}):
    #                 [dq] = (q[uv_i[(u, v)]] - q0[uv_i[(u, v)]])
    #                 form_.edge_attribute((u, v), 'dq', value=dq)

    #             plot_grad(form_).show()

    #         form_, force = paralelise_form(form_, force_, q.flatten(), alpha=alpha)  # , plot=True)
    #         a = max(form.edges_attribute('a'))

    #         f = f_exp(form_)
    #         # print('f = {0:.2f}'.format(f))
    #         if a > a_limit:
    #             f += 10*a
    #             # print('f penalized = {0:.2f}'.format(f))

    #         # qout = array([form_.edge_attribute((u,v),'q') for u, v in form_.edges_where({'_is_edge': True})])[:, newaxis]
    #         # print('q range after grad  : {0:.3f} : {1:.3f}'.format(min(qout), max(qout)))

    #         # plot_form(form_).show()

    #         if f < f0:
    #             break
    #         else:
    #             if step/10 < incr:
    #                 incr = incr/2

    #             step = step - incr

    #             if k % 10 == 0:
    #                 print('Iteration steplength: {0} / Stepsize: {1} / Angle Deviation: {2} / Energy Var: {3:.1f}%'.format(k, step, a, (f-f0)/f0*100))

    #             if k > kmax:
    #                 print('Iterations on steplength reached the maximum {0}'.format(kmax))
    #                 break

    #     # Verify Deviations

    #     # a = evaluate_a(form, print=False)

    #     for u, v in form_.edges_where({'_is_edge': True}):
    #         [dq] = (q[uv_i[(u, v)]] - q0[uv_i[(u, v)]])
    #         form_.edge_attribute((u, v), 'dq', value=dq)

    #     for key, vertex in form_.vertex.items():
    #         i = k_i[key]
    #         z[i] = vertex.get('z')

    #     # Update Function

    #     if f < f0:
    #         form = form_
    #         q = array([form.edge_attribute((u, v), 'q') for u, v in form.edges_where({'_is_edge': True})])[:, newaxis]
    #         # stepsave = step
    #         # incrsave = incr
    #         print('End of Iteration {0} / Step number: {1} / Stepsize: {2:.5f} / Evergy Var: {3:.1f}% / Energy: {4:.1f} / Angle Dev: {5:.3f}'.format(it, k, step,
    #                                                                                                                                                  (f-f0)/f0*100, f, a))
    #         # form = paralelise_form(form,q)

    #     else:
    #         print('Optimisation not found - Iteration {0} / Stepsize: {1:.5f} / Energy: {2:.1f} / Angle Dev: {3}'.format(it, step, f, a))
    #         plot_form(form_).show()
    #         plot_force(force, form).show()
    #         break

    #     if save_steps and it % 10 == 0:
    #         print_ += 1
    #         form.to_json('/Users/mricardo/compas_dev/me/bestfit/iterations/pillow_' + str(print_) + '.json')

    #     it += 1

    #     if plot:
    #         if it % 5 == 0:
    #             plot_form(form).show()
    #             plot_force(force, form).show()

    print('WIP')

    return form


def grad_target(q, z, args):

    C, Ci, Cit, x, y, Wfree, s, free = args

    D = Cit.dot(diags(q.flatten())).dot(C)
    Di = D[:, free]

    # Grad Calculation

    W = diags(C.dot(z).flatten())  # dimension m x m
    J = spsolve(Di, Cit.dot(W))  # dimension ni x m
    grad = ((Wfree * - 2 * (z[free] - s[free])).transpose() * J).transpose()

    return grad


def grad_lp(q, z, args):

    C, Ci, Cit, x, y, Wfree, s, free = args

    D = Cit.dot(diags(q.flatten())).dot(C)
    Di = D[:, free]

    # Grad Calculation

    grad1 = (C.dot(x))**2 + (C.dot(y))**2 + (C.dot(z))**2
    W = diags(C.dot(z).flatten())  # dimension m x m
    J = spsolve(Di, Cit.dot(W))  # dimension ni x m
    M = Ci.dot(J)  # dimension m x m verify
    K = W.dot(M)  # Shape m x 1
    grad2 = (2 * q.transpose() * K).transpose()
    grad = grad1 + grad2

    return grad
